import numpy as np
import scipy.signal as signal
import matplotlib.pyplot as plt
from sortedcontainers import SortedDict
from deepcinac.utils.display import BREWER_COLORS
from deepcinac.utils.utils import get_continous_time_periods
from deepcinac.utils.cinac_file_utils import CinacFileReader
from datetime import datetime
import os
import yaml
from deepcinac.utils.signal import smooth_convolve
import hdf5storage
from scipy.stats import wilcoxon
from matplotlib.patches import Patch, Circle


def build_spike_nums_dur(spike_nums, peak_nums, traces=None, fluorescence_threshold=None):
    """
    Build a "raster dur", meaning a 2d binary array (n_cells * n_frames) representing when a cell is active
    based on the onset and peak on transients. 
    Args:
        spike_nums: (2d array binary n_cells * n_frames), represent the onsets of cell activations (transients)
        peak_nums: (2d array binary n_cells * n_frames), represent the peaks of cell activations (transients)
        traces: (2d array float n_cells * n_frames) fluorescence traces of the cells
        fluorescence_threshold:  if a peak amplitude is under the threshold, we don't consider i

    Returns:

    """
    n_cells = len(spike_nums)
    n_frames = spike_nums.shape[1]
    spike_nums_dur = np.zeros((n_cells, n_frames), dtype="int8")
    for cell in np.arange(n_cells):
        peaks_index = np.where(peak_nums[cell, :])[0]
        onsets_index = np.where(spike_nums[cell, :])[0]

        for onset_index in onsets_index:
            peaks_after = np.where(peaks_index > onset_index)[0]
            if len(peaks_after) == 0:
                continue
            peaks_after = peaks_index[peaks_after]
            peak_after = peaks_after[0]

            if (traces is not None) and (fluorescence_threshold is not None):
                # if the peak amplitude is under the threshold, we don't consider it
                if traces[cell, peak_after] < fluorescence_threshold[cell]:
                    continue

            spike_nums_dur[cell, onset_index:peak_after + 1] = 1
    return spike_nums_dur


def get_raster_dur_from_traces(traces, fluorescence_threshold=None):
    """
    Allow to build a raster dur based on all putative transient from the fluorescence traces.
    :param traces:(2d array float, n_cells *n_frames) fluorescence traces
    :param fluorescence_threshold: None or otherwise 1xlen(traces) array with for each cell the threshold under which
    we should not take into account a peak and the transient associated. The value is without normalization.
    :return:
    """
    n_cells = traces.shape[0]
    n_times = traces.shape[1]

    for i in np.arange(n_cells):
        if fluorescence_threshold is not None:
            fluorescence_threshold[i] = (fluorescence_threshold[i] - np.mean(traces[i, :])) / np.std(traces[i, :])
        traces[i, :] = (traces[i, :] - np.mean(traces[i, :])) / np.std(traces[i, :])

    spike_nums_all = np.zeros((n_cells, n_times), dtype="int8")
    for cell in np.arange(n_cells):
        onsets = []
        diff_values = np.diff(traces[cell])
        for index, value in enumerate(diff_values):
            if index == (len(diff_values) - 1):
                continue
            if value < 0:
                if diff_values[index + 1] >= 0:
                    onsets.append(index + 1)
        if len(onsets) > 0:
            spike_nums_all[cell, np.array(onsets)] = 1

    peak_nums = np.zeros((n_cells, n_times), dtype="int8")
    for cell in np.arange(n_cells):
        peaks, properties = signal.find_peaks(x=traces[cell])
        peak_nums[cell, peaks] = 1

    spike_nums_dur = build_spike_nums_dur(spike_nums_all, peak_nums,
                                          traces=traces, fluorescence_threshold=fluorescence_threshold)
    return spike_nums_dur


def compute_stats_over_gt(raster_gt, raster_to_evaluate, traces, raster_predictions=None, fluorescence_threshold=None):
    """
    Compute the stats based on raster dur
    :param raster_gt: 1d binary array (n_frames or 2d binary array (n_cells, n_frames), 1 when the cell is active
    :param raster_to_evaluate: same sahpe as raster_t, binary, 1 when a cell is active
    :param traces: (1d or 2d array, same shape as raster_gt), should be smoothed, allows to detect
    :raster_predictions: same shape as raster_gt, but float array, probability that the cell is active at each
    frame. Can be None, if ground_truth is not based
    putative transients
    fluorescence_threshold: if not None, float value representing a low threshold for
        # for transients used for benchmarks (possible transients), transients below the threshold are not considered
    :return: two dicts: first one with stats on frames, the other one with stats on transients
    Frames dict has the following keys (as String):
    TP: True Positive
    FP: False Positive
    FN: False Negative
    TN: True Negative
    sensitivity or TPR: True Positive Rate or Recall
    specificity or TNR: True Negative Rate or Selectivity
    FPR: False Positive Rate or Fall-out
    FNR: False Negative Rate or Miss Rate
    ACC: accuracy
    Prevalence: sum positive conditions / total population (for frames only)
    PPV: Positive Predictive Value or Precision
    FDR: False Discovery Rate
    FOR: False Omission Rate
    NPV: Negative Predictive Value
    LR+: Positive Likelihood ratio
    LR-: Negative likelihood ratio

    transients dict has just the following keys:
    TP
    FN
    sensitivity or TPR: True Positive Rate or Recall
    FNR: False Negative Rate or Miss Rate
    """

    if raster_gt.shape != raster_to_evaluate.shape:
        raise Exception(f"both raster_gt and to_evaluate should have the same shape {raster_gt.shape} vs "
                        f"{raster_to_evaluate.shape}")
    if len(raster_gt.shape) == 1:
        # we transform them in a 2 dimensions array
        raster_gt = raster_gt.reshape(1, raster_gt.shape[0])
        raster_to_evaluate = raster_to_evaluate.reshape(1, raster_to_evaluate.shape[0])
        if raster_predictions is not None:
            raster_predictions = raster_predictions.reshape(1, raster_predictions.shape[0])
        if traces is not None:
            traces = traces.reshape(1, traces.shape[0])

    frames_stat = dict()
    transients_stat = dict()

    n_frames = raster_gt.shape[1]
    n_cells = raster_gt.shape[0]

    # full raster dur represents the raster dur built from all potential onsets and peaks
    full_raster_dur = None
    if traces is not None:
        full_raster_dur = get_raster_dur_from_traces(traces, fluorescence_threshold=fluorescence_threshold)

    # positive means active frame, negative means non-active frames
    # condition is the ground truth
    # predicted is the one computed (RNN, CaiMan etc...)

    tp_frames = 0
    fp_frames = 0
    fn_frames = 0
    tn_frames = 0

    tp_transients = 0
    fn_transients = 0
    fp_transients = 0
    tn_transients = 0

    # will keep values of the predictions of the FN and FP transients and frames
    # for transients will take median predicted value over the transient
    fn_transients_predictions = []
    fp_transients_predictions = []
    tp_transients_predictions = []
    tn_transients_predictions = []
    fn_frames_predictions = []
    fp_frames_predictions = []
    tn_frames_predictions = []
    tp_frames_predictions = []

    proportion_of_frames_detected_in_transients = []

    for cell in np.arange(n_cells):
        raster_dur = raster_gt[cell]
        predicted_raster_dur = raster_to_evaluate[cell]
        if raster_predictions is not None:
            gt_predictions_for_cell = raster_predictions[cell]
        else:
            gt_predictions_for_cell = None
        predicted_positive_frames = np.where(predicted_raster_dur)[0]
        predicted_negative_frames = np.where(predicted_raster_dur == 0)[0]

        tp_frames += len(np.where(raster_dur[predicted_positive_frames] == 1)[0])
        fp_frames += len(np.where(raster_dur[predicted_positive_frames] == 0)[0])
        fn_frames += len(np.where(raster_dur[predicted_negative_frames] == 1)[0])
        tn_frames += len(np.where(raster_dur[predicted_negative_frames] == 0)[0])

        if raster_predictions is not None:
            fp_frames_indices = np.where(raster_dur[gt_predictions_for_cell >= 0.5] == 0)[0]
            fn_frames_indices = np.where(raster_dur[gt_predictions_for_cell < 0.5] == 1)[0]
            tp_frames_indices = np.where(raster_dur[gt_predictions_for_cell >= 0.5] == 1)[0]
            tn_frames_indices = np.where(raster_dur[gt_predictions_for_cell < 0.5] == 0)[0]
            fp_frames_predictions.extend(list(gt_predictions_for_cell[fp_frames_indices]))
            fn_frames_predictions.extend(list(gt_predictions_for_cell[fn_frames_indices]))
            tp_frames_predictions.extend(list(gt_predictions_for_cell[tp_frames_indices]))
            tn_frames_predictions.extend(list(gt_predictions_for_cell[tn_frames_indices]))

        n_fake_transients = 0
        # transients section
        transient_periods = get_continous_time_periods(raster_dur)
        if full_raster_dur is not None:
            full_transient_periods = get_continous_time_periods(full_raster_dur[cell])
            fake_transients_periods = []
            # keeping only the fake ones
            for transient_period in full_transient_periods:
                if np.sum(raster_dur[transient_period[0]:transient_period[1] + 1]) > 0:
                    # it means it's a real transient
                    continue
                fake_transients_periods.append(transient_period)

            n_fake_transients = len(fake_transients_periods)
        # positive condition
        n_transients = len(transient_periods)
        tp = 0
        for transient_period in transient_periods:
            frames = np.arange(transient_period[0], transient_period[1] + 1)
            if np.sum(predicted_raster_dur[frames]) > 0:
                tp += 1
                # keeping only transients with one frame detected
                proportion_of_frames_detected_in_transients.append((np.sum(predicted_raster_dur[frames]) / len(frames))
                                                                   * 100)

            if raster_predictions is not None:
                if np.max(gt_predictions_for_cell[frames]) >= 0.5:
                    # adding the median of the predicted frames in the transient
                    tp_transients_predictions.append(np.max(gt_predictions_for_cell[frames]))
                else:
                    # then if's a FN
                    fn_transients_predictions.append(np.max(gt_predictions_for_cell[frames]))
        tn = 0
        if full_raster_dur is not None:
            for transient_period in fake_transients_periods:
                frames = np.arange(transient_period[0], transient_period[1] + 1)
                if np.sum(predicted_raster_dur[frames]) == 0:
                    tn += 1
                if raster_predictions is not None:
                    if np.max(gt_predictions_for_cell[frames]) < 0.5:
                        # adding the max of the predicted frames in the transient
                        # by max we know how far we are from 0.5
                        tn_transients_predictions.append(np.max(gt_predictions_for_cell[frames]))
                    else:
                        # then if's a FP
                        # taking the max because we want to see how far are we from 0.5
                        fp_transients_predictions.append(np.max(gt_predictions_for_cell[frames]))

        tp_transients += tp
        fn_transients += (n_transients - tp)
        tn_transients += tn
        fp_transients += (n_fake_transients - tn)

    frames_stat["TP"] = tp_frames
    frames_stat["FP"] = fp_frames
    frames_stat["FN"] = fn_frames
    frames_stat["TN"] = tn_frames

    # frames_stat["TPR"] = tp_frames / (tp_frames + fn_frames)
    if (tp_frames + fn_frames) > 0:
        frames_stat["sensitivity"] = tp_frames / (tp_frames + fn_frames)
    else:
        frames_stat["sensitivity"] = 1
    frames_stat["TPR"] = frames_stat["sensitivity"]

    if (tn_frames + fp_frames) > 0:
        frames_stat["specificity"] = tn_frames / (tn_frames + fp_frames)
    else:
        frames_stat["specificity"] = 1
    frames_stat["TNR"] = frames_stat["specificity"]

    if (tp_frames + tn_frames + fp_frames + fn_frames) > 0:
        frames_stat["ACC"] = (tp_frames + tn_frames) / (tp_frames + tn_frames + fp_frames + fn_frames)
    else:
        frames_stat["ACC"] = 1

    if (tp_frames + fp_frames) > 0:
        frames_stat["PPV"] = tp_frames / (tp_frames + fp_frames)
    else:
        frames_stat["PPV"] = 1
    if (tn_frames + fn_frames) > 0:
        frames_stat["NPV"] = tn_frames / (tn_frames + fn_frames)
    else:
        frames_stat["NPV"] = 1

    frames_stat["FNR"] = 1 - frames_stat["TPR"]

    frames_stat["FPR"] = 1 - frames_stat["TNR"]

    if "PPV" in frames_stat:
        frames_stat["FDR"] = 1 - frames_stat["PPV"]

    if "NPV" in frames_stat:
        frames_stat["FOR"] = 1 - frames_stat["NPV"]

    if frames_stat["FPR"] > 0:
        frames_stat["LR+"] = frames_stat["TPR"] / frames_stat["FPR"]
    else:
        frames_stat["LR+"] = 1

    if frames_stat["TNR"] > 0:
        frames_stat["LR-"] = frames_stat["FNR"] / frames_stat["TNR"]
    else:
        frames_stat["LR-"] = 1

    # transients dict
    transients_stat["TP"] = tp_transients
    transients_stat["FN"] = fn_transients
    if traces is not None:
        # print(f"tn_transients {tn_transients}")
        transients_stat["TN"] = tn_transients
        transients_stat["FP"] = fp_transients

    if (tp_transients + fn_transients) > 0:
        transients_stat["sensitivity"] = tp_transients / (tp_transients + fn_transients)
    else:
        transients_stat["sensitivity"] = 1

    # print(f'transients_stat["sensitivity"] {transients_stat["sensitivity"]}')
    transients_stat["TPR"] = transients_stat["sensitivity"]

    if traces is not None:
        if (tn_transients + fp_transients) > 0:
            transients_stat["specificity"] = tn_transients / (tn_transients + fp_transients)
        else:
            transients_stat["specificity"] = 1
        transients_stat["TNR"] = transients_stat["specificity"]

        if (tp_transients + tn_transients + fp_transients + fn_transients) > 0:
            transients_stat["ACC"] = (tp_transients + tn_transients) / \
                                     (tp_transients + tn_transients + fp_transients + fn_transients)
        else:
            transients_stat["ACC"] = 1

        if (tp_transients + fp_transients) > 0:
            transients_stat["PPV"] = tp_transients / (tp_transients + fp_transients)
        else:
            transients_stat["PPV"] = 1
        if (tn_transients + fn_transients) > 0:
            transients_stat["NPV"] = tn_transients / (tn_transients + fn_transients)
        else:
            transients_stat["NPV"] = 1

    transients_stat["FNR"] = 1 - transients_stat["TPR"]

    if raster_predictions is not None:
        predictions_stat = dict()
        predictions_stat["fn_transients_predictions"] = fn_transients_predictions
        predictions_stat["fp_transients_predictions"] = fp_transients_predictions
        predictions_stat["tp_transients_predictions"] = tp_transients_predictions
        predictions_stat["tn_transients_predictions"] = tn_transients_predictions
        predictions_stat["fn_frames_predictions"] = fn_frames_predictions
        predictions_stat["fp_frames_predictions"] = fp_frames_predictions
        predictions_stat["tn_frames_predictions"] = tn_frames_predictions
        predictions_stat["tp_frames_predictions"] = tp_frames_predictions
    else:
        predictions_stat = None

    return frames_stat, transients_stat, predictions_stat, proportion_of_frames_detected_in_transients


class CinacBenchmarks:
    """
    Used to plot benchmarks.
    To do so 4 steps:
    Create an instance of CinacBenchmarks
    Add data using add_inference_to_benchmark()
    Add color for session if you want to using color_by_session()
    then call one the method to plot stats
    """

    def __init__(self, results_path, colors_boxplots=None, verbose=0):
        # key is a session_id (str), value is a 2d binary array (n_cells * n_frames)
        self.ground_truth_dict = dict()
        # key is a session_id (str), value is a 2d array (n_cells * n_frames)
        self.smooth_traces_dict = dict()
        self.verbose = verbose
        self.results_path = results_path

        if colors_boxplots is None:
            self.colors_boxplots = ['#a6cee3', '#1f78b4', '#b2df8a', '#33a02c', '#fb9a99', '#e31a1c',
                                    '#ff7f00', '#cab2d6', '#fdbf6f', '#6a3d9a', '#ffff99', '#b15928', '#ffffd9']
        else:
            self.colors_boxplots = colors_boxplots

        # self.colors_boxplots = self.colors_boxplots[::-1]

        # key is reference_id and value is a boolean, indicating if predictions are available for this inference
        # meaning values that give the probability for the cell to be active at a given frame
        self.are_predictions_for_inference = dict()

        # dict with key the id (str) of the inference method,
        # and value is list of one of 2 array of same shape as ground_truth (1d array), first array is a binary one,
        # 2nd one (optional) is a float one representing the probability (between 0 and 1) for the cell to be active
        self.inferences_to_benchmark_dict = dict()

        # key are the same as inferences_to_benchmark_dict, value is a list of int representing the cell indices
        # to benchmark
        self.cells_to_benchmark_dict = dict()

        # inference id-> session_id -> cell -> value will be a list of dict with results from benchmarks
        self.results_frames_dict = dict()
        self.results_transients_dict = dict()
        # inference id-> session_id-> cell -> value is a list of percentage
        # that represents for each real transient the proportion of frames detected
        self.results_frames_in_transients_pc = dict()
        # inference id-> session_id-> cell -> value will be a dict with results from benchmarks. Each key of the dict
        # each key being a string like "fp_transients_predictions" or "fp_frames_predictions"
        self.results_predictions_dict = dict()
        # inference id-> session_id> value will be a list of dict with results from benchmarks
        self.results_dict_global = dict()

        # key: session_id, value: color
        self.cells_to_color = dict()

    def color_by_session(self, session_id, color):
        """
        Attribute a cell to a color, allowing for example to give to cell of the same
        Args:
            session_id: (str)
            color:

        Returns:

        """
        self.cells_to_color[session_id] = color

    def add_ground_truth(self, session_id, ground_truth, smooth_traces):
        """

        Args:
            session_id:
            ground_truth:
            smooth_traces

        Returns:

        """
        self.ground_truth_dict[session_id] = ground_truth
        self.smooth_traces_dict[session_id] = smooth_traces

    def add_inference_to_benchmark(self, session_id, inference_to_benchmark_id, raster_to_evaluate,
                                   cells_to_benchmark,
                                   raster_predictions=None):
        """

        Args:
            session_id:
            inference_to_benchmark_id:
            raster_to_evaluate:
            raster_predictions:
            cells_to_benchmark:

        Returns:

        """

        if inference_to_benchmark_id not in self.cells_to_benchmark_dict:
            self.cells_to_benchmark_dict[inference_to_benchmark_id] = dict()
        if session_id in self.cells_to_benchmark_dict[inference_to_benchmark_id]:
            print(f"Session {session_id} for {inference_to_benchmark_id} already added to inference_to_benchmark")

        self.cells_to_benchmark_dict[inference_to_benchmark_id][session_id] = cells_to_benchmark
        inference_data = [raster_to_evaluate]
        if raster_predictions is not None:
            inference_data.append(raster_predictions)

        # assuming that if a predictions is available for an inference method, it is for all sessions
        if inference_to_benchmark_id not in self.are_predictions_for_inference:
            self.are_predictions_for_inference[inference_to_benchmark_id] = False
        if raster_predictions is not None:
            self.are_predictions_for_inference[inference_to_benchmark_id] = True

        if inference_to_benchmark_id not in self.inferences_to_benchmark_dict:
            self.inferences_to_benchmark_dict[inference_to_benchmark_id] = dict()
        self.inferences_to_benchmark_dict[inference_to_benchmark_id][session_id] = inference_data

        if inference_to_benchmark_id not in self.results_predictions_dict:
            self.results_predictions_dict[inference_to_benchmark_id] = dict()
        self.results_predictions_dict[inference_to_benchmark_id][session_id] = SortedDict()

        if inference_to_benchmark_id not in self.results_frames_dict:
            self.results_frames_dict[inference_to_benchmark_id] = dict()
        self.results_frames_dict[inference_to_benchmark_id][session_id] = SortedDict()

        if inference_to_benchmark_id not in self.results_transients_dict:
            self.results_transients_dict[inference_to_benchmark_id] = dict()
        self.results_transients_dict[inference_to_benchmark_id][session_id] = SortedDict()

        if inference_to_benchmark_id not in self.results_frames_in_transients_pc:
            self.results_frames_in_transients_pc[inference_to_benchmark_id] = dict()
        self.results_frames_in_transients_pc[inference_to_benchmark_id][session_id] = SortedDict()

    def evaluate_metrics(self):
        """
        Should be called after all inference to benchmark have been added and before generating the plots
        Returns:

        """
        for inference_to_benchmark_id, cell_dict in self.cells_to_benchmark_dict.items():
            for session_id, cells in cell_dict.items():
                for cell in cells:
                    self._evaluate_metrics_on_a_cell(session_id=session_id,
                                                     inference_to_benchmark_id=inference_to_benchmark_id,
                                                     cell=cell)

    def _evaluate_metrics_on_a_cell(self, session_id, inference_to_benchmark_id, cell):
        """

        Args:
            session_id: session from which the activity is recorded
            inference_to_benchmark_id: (str) id of the inference to benchmark
            cell: int

        Returns: a dict with same keys as inferences_to_benchmarks and values...
        """

        if inference_to_benchmark_id not in self.inferences_to_benchmark_dict:
            print(f"evaluate_metrics_on_a_cell: inference {inference_to_benchmark_id} unknown")

        if session_id not in self.inferences_to_benchmark_dict[inference_to_benchmark_id]:
            print(f"evaluate_metrics_on_a_cell: session {session_id} unknown")
            return

        inference_to_benchmark = self.inferences_to_benchmark_dict[inference_to_benchmark_id][session_id]
        # print(f"{session_id} {inference_to_benchmark_id}: n_cells for evaluation {len(inference_to_benchmark[0])}")
        raster_to_evaluate = inference_to_benchmark[0][cell]
        if len(inference_to_benchmark) == 2:
            raster_predictions = inference_to_benchmark[1][cell]
        else:
            raster_predictions = None

        frames_stat, transients_stat, predictions_stat_dict, proportion_of_frames_detected_in_transients = \
            compute_stats_over_gt(raster_gt=self.ground_truth_dict[session_id][cell],
                                  raster_predictions=raster_predictions,
                                  raster_to_evaluate=raster_to_evaluate,
                                  traces=self.smooth_traces_dict[session_id][cell],
                                  fluorescence_threshold=None)

        self.results_predictions_dict[inference_to_benchmark_id][session_id][cell] = predictions_stat_dict
        self.results_frames_dict[inference_to_benchmark_id][session_id][cell] = frames_stat
        self.results_transients_dict[inference_to_benchmark_id][session_id][cell] = transients_stat
        self.results_frames_in_transients_pc[inference_to_benchmark_id][session_id][
            cell] = proportion_of_frames_detected_in_transients

        if self.verbose:
            # frames stats
            print(f"raster {inference_to_benchmark_id}")
            print(f"Frames stat:")
            for k, value in frames_stat.items():
                print(f"{k}: {str(np.round(value, 4))}")
        if self.verbose:
            print(f"###")
            print(f"Transients stat:")
            for k, value in transients_stat.items():
                print(f"{k}: {str(np.round(value, 4))}")
            print("")

    def plot_boxplot_predictions_stat_by_metrics(self, description, time_str=None, colorfull=True,
                                                 white_background=True, for_frames=False,
                                                 save_formats="pdf", dpi=500):
        """
        PLot the boxplot regarding the predictions done by cinac classifiers.
        Args:
            description: (str) will be added to file_name
            time_str: (str) timestamps for file_name, optional
            save_formats:str or list of str (.pdf, .png etc...)
            dpi:

        Returns:

        """

        if len(self.results_predictions_dict) == 0:
            # it means no inference had predictions
            return

        if for_frames:
            type_of_activity_list = ["frames"]
        else:
            type_of_activity_list = ["transients"]
        metrics_to_show_list = ["tp", "tn", "fp", "fn"]

        with_scatter = True

        for inference_id in list(self.results_predictions_dict.keys()):
            # first we check if we have predictions for this inference
            if not self.are_predictions_for_inference[inference_id]:
                continue
            # print(f"Inference {inference_id} in the loop")
            """
            # inference id-> session_id-> cell -> value will be a dict with results from benchmarks. Each key of the dict
        # each key being a string like "fp_transients_predictions" or "fp_frames_predictions"
        self.results_predictions_dict = dict()
            """
            predictions_by_cell_stat = dict()
            predictions_stat = dict()
            # we take just the first data, as it's supposed to be same results for each
            for type_of_activity in type_of_activity_list:
                predictions_by_cell_stat[type_of_activity] = dict()
                predictions_stat[type_of_activity] = dict()
                for metrics_to_show in metrics_to_show_list:
                    predictions_by_cell_stat[type_of_activity][metrics_to_show] = dict()
                    predictions_stat[type_of_activity][metrics_to_show] = []
                    data_pred_dict = self.results_predictions_dict[inference_id]
                    for session_id, session_data_dict in data_pred_dict.items():
                        predictions_by_cell_stat[type_of_activity][metrics_to_show][session_id] = dict()
                        for cell, pred_dict in session_data_dict.items():
                            string_key = metrics_to_show + "_" + type_of_activity + "_" + "predictions"
                            predictions_by_cell_stat[type_of_activity][metrics_to_show][session_id][cell] = \
                                pred_dict[string_key]
                            predictions_stat[type_of_activity][metrics_to_show].extend(pred_dict[string_key])
                            # first one is enough

            for by_cell in [True, False]:
                for type_of_activity in type_of_activity_list:
                    if by_cell:
                        # using the value in metrics_to_show
                        n_box_plots = 0
                        for session_id, session_data_dict in \
                                predictions_by_cell_stat[type_of_activity][metrics_to_show].items():
                            # number of cells in the session
                            n_box_plots += len(session_data_dict)
                        if n_box_plots > 25:
                            # if more than 15 boxplots, we don't plot it
                            continue
                    stat_fig, axes = plt.subplots(nrows=2, ncols=2, squeeze=True,
                                                  gridspec_kw={'height_ratios': [0.5, 0.5],
                                                               'width_ratios': [0.5, 0.5]},
                                                  figsize=(10, 10), dpi=dpi)

                    stat_fig.set_tight_layout({'rect': [0, 0, 1, 1], 'pad': 1, 'w_pad': 1, 'h_pad': 5})
                    axes = np.ndarray.flatten(axes)
                    fig_patch = stat_fig.patch

                    # rgba = c_map(0)
                    if white_background:
                        face_color = "white"
                        text_color = "black"
                        title_color = "black"
                    else:
                        face_color = "black"
                        text_color = "white"
                        title_color = "red"
                    fig_patch.set_facecolor(face_color)

                    for metrics_index, metrics_to_show in enumerate(metrics_to_show_list):
                        ax = axes[metrics_index]

                        ax.set_facecolor(face_color)

                        ax.set_frame_on(False)
                        box_plot_index_so_far = 0
                        if by_cell:
                            n_box_plots = 0
                            labels = []
                            values_by_prediction = []
                            for session_id, by_cell_dict in \
                                    predictions_by_cell_stat[type_of_activity][metrics_to_show].items():
                                n_box_plots += len(by_cell_dict)
                                for cell_key in by_cell_dict.keys():
                                    labels.append(f"{len(by_cell_dict[cell_key])}")
                                # labels = list(by_cell_dict.keys())
                                values_by_prediction.extend([by_cell_dict[cell_key] for cell_key in
                                                             by_cell_dict.keys()])
                                if with_scatter:
                                    for label_index, cell_key in enumerate(list(by_cell_dict.keys())):
                                        # print(f"scatter {label_index} {cell_key}")
                                        y_pos = by_cell_dict[cell_key]
                                        x_pos = []
                                        for ii in range(len(y_pos)):
                                            # Adding jitter
                                            x_pos.append(
                                                1 + box_plot_index_so_far + ((np.random.random_sample() - 0.5) * 0.7))
                                        if white_background:
                                            edgecolors = "black"
                                        else:
                                            edgecolors = "white"
                                        ax.scatter(x_pos, y_pos,
                                                   color=BREWER_COLORS[box_plot_index_so_far % len(BREWER_COLORS)],
                                                   marker="o",
                                                   edgecolors=edgecolors,
                                                   s=20, zorder=21, alpha=0.4)
                                        box_plot_index_so_far += 1
                        else:
                            n_cells = 0
                            for session_id, by_cell_dict in \
                                    predictions_by_cell_stat[type_of_activity][metrics_to_show].items():
                                n_cells += len(by_cell_dict)
                            labels = [f"{n_cells} cells, "
                                      f"{len(predictions_stat[type_of_activity][metrics_to_show])} {type_of_activity}"]
                            n_box_plots = 1
                            values_by_prediction = [predictions_stat[type_of_activity][metrics_to_show]]

                            if with_scatter:
                                y_pos = predictions_stat[type_of_activity][metrics_to_show]
                                x_pos = []
                                for ii in range(len(y_pos)):
                                    # Adding jitter
                                    x_pos.append(1 + ((np.random.random_sample() - 0.5) * 0.7))
                                if white_background:
                                    edgecolors = "black"
                                else:
                                    edgecolors = "white"
                                ax.scatter(x_pos, y_pos,
                                           color=BREWER_COLORS[0 % len(BREWER_COLORS)],
                                           marker="o",
                                           edgecolors=edgecolors,
                                           s=20, zorder=21, alpha=0.4)

                        outliers = dict(markerfacecolor='white', marker='D')
                        # if not for_frames:
                        #     print(f"plot_boxplots_full_stat: {stat_name}: values_by_prediction {values_by_prediction}")
                        if with_scatter:
                            sym = ''
                        else:
                            sym = '+'

                        # print("stat_by_metrics")
                        # print(f"{metrics_to_show.upper()}: median: {np.round(np.median(values_by_prediction), 3)}, "
                        #       f"25p {np.round(np.percentile(values_by_prediction, 25), 3)}, "
                        #       f"75p {np.round(np.percentile(values_by_prediction, 75), 3)}")

                        bplot = ax.boxplot(values_by_prediction, patch_artist=colorfull,
                                           flierprops=outliers, widths=[0.7] * len(values_by_prediction),
                                           labels=labels, sym=sym, zorder=1)  # whis=[5, 95], sym='+'

                        for element in ['boxes', 'whiskers', 'fliers', 'caps']:
                            if white_background:
                                plt.setp(bplot[element], color="black")
                            else:
                                plt.setp(bplot[element], color="white")

                        for element in ['means', 'medians']:
                            if white_background:
                                plt.setp(bplot[element], color="black")
                            else:
                                plt.setp(bplot[element], color="white")

                        if colorfull:
                            colors = self.colors_boxplots[:n_box_plots]
                            for patch, color in zip(bplot['boxes'], colors):
                                patch.set_facecolor(color)

                        ax.xaxis.set_ticks_position('none')
                        ax.xaxis.label.set_color(text_color)
                        ax.tick_params(axis='x', colors=text_color)
                        if n_box_plots <= 2:
                            ax.xaxis.set_tick_params(labelsize=15)
                        elif n_box_plots <= 6:
                            ax.xaxis.set_tick_params(labelsize=9)
                        else:
                            ax.xaxis.set_tick_params(labelsize=6)
                        ax.yaxis.label.set_color(text_color)
                        ax.tick_params(axis='y', colors=text_color)
                        # ax.set_xticklabels([])
                        # ax.set_yticklabels([])
                        # ax.get_yaxis().set_visible(False)
                        # ax.get_xaxis().set_visible(False)
                        # ax.set_ylabel(f"proportion")
                        # ax.set_xlabel("age")
                        xticks = np.arange(1, n_box_plots + 1)
                        ax.set_xticks(xticks)
                        # sce clusters labels
                        ax.set_xticklabels(labels)
                        # fixing the limits
                        # if stat_name == "sensitivity":
                        #     ax.set_ylim(0, 1)
                        # elif stat_name == "specificity":
                        #     ax.set_ylim(0.85, 1)
                        # elif stat_name == "PPV":
                        #     ax.set_ylim(0, 1)
                        # elif stat_name == "NPV":
                        #     ax.set_ylim(0.6, 1.1)

                        ax.set_title(metrics_to_show.upper(), color=title_color, pad=20, fontsize=30, fontweight='bold')

                    str_details = "and_cells_" if by_cell else ""
                    str_details = str_details + "_" + inference_id
                    if isinstance(save_formats, str):
                        save_formats = [save_formats]
                    if time_str is None:
                        time_str = datetime.now().strftime("%Y_%m_%d.%H-%M-%S")
                    for save_format in save_formats:
                        stat_fig.savefig(f'{self.results_path}/'
                                         f'{description}_box_plots_predictions_by_metrics_{str_details}_'
                                         f'for_{type_of_activity}'
                                         f'_{time_str}.{save_format}',
                                         format=f"{save_format}",
                                         facecolor=stat_fig.get_facecolor(), edgecolor='none')
                    plt.close()

    def plot_boxplots_full_stat(self, description, time_str=None, for_frames=True, with_cells=False,
                                color_cell_as_boxplot=False,
                                box_plots_labels=None,
                                colorfull=True,
                                white_background=True,
                                stats_to_show=("sensitivity", "specificity", "PPV", "NPV"),
                                title_correspondance=None,
                                alpha_scatter=1., with_cell_number=True,
                                put_metric_as_y_axis_label=False,
                                save_formats="pdf", dpi=500):
        """

        :param description:
        :param time_str:
        :param for_frames:
        ;param box_plots_labels: if not NOne, list of str, allowing to choose the order of the boxplots.
        The labels should be existing one, otherwise it will crash
        :param with_cells: if True, display a scatter for each cell
        :param save_formats:
        :param title_correspondance: dict or None, keys are the stats to show, and the value is another string
        that will be displayed in the title
        :return:
        """
        result_dict_to_use = self.results_frames_dict
        if not for_frames:
            result_dict_to_use = self.results_transients_dict

        nrows = 2
        height_ratios = [0.5, 0.5]
        ncols = 2
        width_ratios = [0.5, 0.5]
        figsize = (10, 10)
        if len(stats_to_show) == 2:
            nrows = 2
            height_ratios = [0.5, 0.5]
            ncols = 1
            width_ratios = [1]
            figsize = (5, 10)
        if len(stats_to_show) == 1:
            nrows = 1
            height_ratios = [1]
            ncols = 1
            width_ratios = [1]
            figsize = (5, 5)
        stat_fig, axes = plt.subplots(nrows=nrows, ncols=ncols, squeeze=True,
                                      gridspec_kw={'height_ratios': height_ratios,
                                                   'width_ratios': width_ratios},
                                      figsize=figsize, dpi=dpi)

        stat_fig.set_tight_layout({'rect': [0, 0, 1, 1], 'pad': 1, 'w_pad': 1, 'h_pad': 5})
        if isinstance(axes, np.ndarray):
            axes = np.ndarray.flatten(axes)
        else:
            axes = [axes]
        fig_patch = stat_fig.patch

        # rgba = c_map(0)
        if white_background:
            face_color = "white"
            text_color = "black"
            title_color = "black"
        else:
            face_color = "black"
            text_color = "white"
            title_color = "red"
        fig_patch.set_facecolor(face_color)

        for stat_index, stat_name in enumerate(stats_to_show):
            ax = axes[stat_index]
            # n_cells = len(self.results_frames_dict_by_cell)

            ax.set_facecolor(face_color)

            ax.set_frame_on(False)
            n_box_plots = None
            labels = None
            values_by_prediction = None

            n_box_plots = len(result_dict_to_use)
            if box_plots_labels is None:
                labels = list(result_dict_to_use.keys())
            else:
                labels = box_plots_labels
            values_by_prediction = [[] for n in np.arange(n_box_plots)]

            for label_index, label in enumerate(labels):
                session_dict = result_dict_to_use[label]
                for session_id, cells_dict in session_dict.items():
                    for cell_to_display, stat_dict in cells_dict.items():
                        # all label might not have the same cells
                        values_by_prediction[label_index].append(stat_dict[stat_name])
                        if with_cells:
                            # Adding jitter
                            x_pos = 1 + label_index + ((np.random.random_sample() - 0.5) * 0.7)
                            y_pos = stat_dict[stat_name]
                            font_size = 3
                            if white_background:
                                edgecolors = "black"
                            else:
                                edgecolors = "white"
                            if color_cell_as_boxplot:
                                color_cell = self.colors_boxplots[label_index % len(self.colors_boxplots)]
                            elif session_id not in self.cells_to_color:
                                color_cell = BREWER_COLORS[label_index % len(BREWER_COLORS)]
                            else:
                                color_cell = self.cells_to_color[session_id]
                            ax.scatter(x_pos, y_pos,
                                       color=color_cell,
                                       marker="o",
                                       edgecolors=edgecolors,
                                       s=60, zorder=21, alpha=alpha_scatter)
                            if cell_to_display > 999:
                                font_size = 2
                            if with_cell_number:
                                ax.text(x=x_pos, y=y_pos,
                                        s=f"{cell_to_display}", color="black", zorder=22,
                                        ha='center', va="center", fontsize=font_size, fontweight='bold')

            outliers = dict(markerfacecolor='white', marker='D')

            print(f"full_stats for_frames {for_frames}, stat {stat_name}: {description}")
            for label_index, label in enumerate(labels):
                print(f"{label}: median: {np.round(np.median(values_by_prediction[label_index]), 3)}, "
                      f"25p {np.round(np.percentile(values_by_prediction[label_index], 25), 3)}, "
                      f"75p {np.round(np.percentile(values_by_prediction[label_index], 75), 3)}")

            bplot = ax.boxplot(values_by_prediction, patch_artist=colorfull,
                               flierprops=outliers, widths=[0.7] * len(values_by_prediction),
                               labels=labels, sym='', zorder=1)  # whis=[5, 95], sym='+'

            for element in ['boxes', 'whiskers', 'fliers', 'caps']:
                if white_background:
                    plt.setp(bplot[element], color="black")
                else:
                    plt.setp(bplot[element], color="white")

            for element in ['means', 'medians']:
                if white_background:
                    plt.setp(bplot[element], color="black")
                else:
                    plt.setp(bplot[element], color="white")

            if colorfull:
                # supposing there are less box_plots than colors,
                # but most probably there won't be more than 20 boxplots
                colors = self.colors_boxplots[:n_box_plots]
                for patch, color in zip(bplot['boxes'], colors):
                    patch.set_facecolor(color)

            ax.xaxis.set_ticks_position('none')
            ax.xaxis.label.set_color(text_color)
            ax.tick_params(axis='x', colors=text_color)
            if n_box_plots <= 2:
                ax.xaxis.set_tick_params(labelsize=20)
            elif n_box_plots <= 4:
                ax.xaxis.set_tick_params(labelsize=15)
            else:
                ax.xaxis.set_tick_params(labelsize=2)
            ax.yaxis.label.set_color(text_color)
            ax.tick_params(axis='y', colors=text_color)
            # ax.set_xticklabels([])
            # ax.set_yticklabels([])
            # ax.get_yaxis().set_visible(False)
            # ax.get_xaxis().set_visible(False)
            if put_metric_as_y_axis_label:
                if title_correspondance is not None and stat_name in title_correspondance:
                    ax.set_ylabel(title_correspondance[stat_name] + " (%)")
                else:
                    ax.set_ylabel(stat_name + " (%)")
                ax.yaxis.label.set_size(20)
            else:
                if title_correspondance is not None and stat_name in title_correspondance:
                    ax.set_title(title_correspondance[stat_name], color=title_color, pad=20, fontsize=20)
                else:
                    ax.set_title(stat_name, color=title_color, pad=20, fontsize=20)

            # ax.set_xlabel("age")
            xticks = np.arange(1, n_box_plots + 1)
            ax.set_xticks(xticks)
            ax.set_xticklabels(labels)

            yticks = np.arange(0, 1.2, 0.2)
            ax.set_yticks(yticks)
            ax.set_yticklabels(np.arange(0, 120, 20))
            # fixing the limits
            if stat_name == "sensitivity":
                ax.set_ylim(0, 1.1)
            # elif stat_name == "specificity":
            #     ax.set_ylim(0.85, 1)
            if stat_name == "PPV":
                ax.set_ylim(0, 1.1)
            # elif stat_name == "NPV":
            #     ax.set_ylim(0.6, 1.1)

        str_details = "frames"
        if not for_frames:
            str_details = "transients"
        for stat_name in stats_to_show:
            str_details = str_details + f'_{title_correspondance[stat_name]}'
        if isinstance(save_formats, str):
            save_formats = [save_formats]
        if time_str is None:
            time_str = datetime.now().strftime("%Y_%m_%d.%H-%M-%S")
        for save_format in save_formats:
            stat_fig.savefig(f'{self.results_path}/'
                             f'{description}_box_plots_predictions_{str_details}'
                             f'_{time_str}.{save_format}',
                             format=f"{save_format}",
                             facecolor=stat_fig.get_facecolor(), edgecolor='none')
        plt.close()

    # TODO: update this one
    def plot_boxplots_for_transients_stat(self, description, colorfull=True, time_str=None, save_formats="pdf"):
        stats_to_show = ["sensitivity"]
        colors = ["cornflowerblue", "blue", "steelblue", "red", "orange"]

        stat_fig, axes = plt.subplots(nrows=1, ncols=1, squeeze=True,
                                      gridspec_kw={'height_ratios': [1],
                                                   'width_ratios': [1]},
                                      figsize=(8, 8))

        stat_fig.set_tight_layout({'rect': [0, 0, 1, 1], 'pad': 1, 'w_pad': 1, 'h_pad': 5})
        # axes = np.ndarray.flatten(axes)
        fig_patch = stat_fig.patch
        # rgba = c_map(0)
        face_color = "black"
        text_color = "white"
        title_color = "red"
        fig_patch.set_facecolor(face_color)

        for stat_index, stat_name in enumerate(stats_to_show):
            ax = axes
            # n_cells = len(self.results_transients_dict_by_cell)

            # now adding as many suplots as need, depending on how many overlap has the cell

            ax.set_facecolor(face_color)
            ax.xaxis.set_ticks_position('none')
            ax.xaxis.label.set_color(text_color)
            ax.tick_params(axis='x', colors=text_color)
            ax.yaxis.label.set_color(text_color)
            ax.tick_params(axis='y', colors=text_color)
            # ax.set_xticklabels([])
            # ax.set_yticklabels([])
            # ax.get_yaxis().set_visible(False)
            # ax.get_xaxis().set_visible(False)

            ax.set_frame_on(False)
            n_box_plots = None
            labels = None
            values_by_prediction = None

            n_box_plots = len(self.results_transients_dict)
            labels = list(self.results_transients_dict.keys())
            values_by_prediction = [[] for n in np.arange(n_box_plots)]
            for label_index, label in enumerate(labels):
                session_dict = self.results_transients_dict[label]
                for session_id, cells_dict in session_dict.items():
                    for cell_to_display, stat_dict in cells_dict.items():
                        values_by_prediction[label_index]. \
                            append(self.results_transients_dict[cell_to_display][label][stat_name])

            bplot = ax.boxplot(values_by_prediction, patch_artist=colorfull,
                               labels=labels, sym='', zorder=1)  # whis=[5, 95], sym='+'

            for element in ['boxes', 'whiskers', 'fliers', 'caps']:
                plt.setp(bplot[element], color="white")

            for element in ['means', 'medians']:
                plt.setp(bplot[element], color="white")  # used to be silver

            if colorfull:
                colors = colors[:n_box_plots]
                for patch, color in zip(bplot['boxes'], colors):
                    patch.set_facecolor(color)

            # ax.set_ylabel(f"proportion")
            # ax.set_xlabel("age")
            xticks = np.arange(1, n_box_plots + 1)
            ax.set_xticks(xticks)
            # sce clusters labels
            ax.set_xticklabels(labels)

            ax.set_title(stat_name, color=title_color, pad=20)

        if time_str is None:
            time_str = datetime.now().strftime("%Y_%m_%d.%H-%M-%S")
        if isinstance(save_formats, str):
            save_formats = [save_formats]
        for save_format in save_formats:
            stat_fig.savefig(f'{self.results_path}/'
                             f'{description}_box_plots_predictions_transients'
                             f'_{time_str}.{save_format}',
                             format=f"{save_format}",
                             facecolor=stat_fig.get_facecolor(), edgecolor='none')

    def stats_on_performance(self, stats_to_evaluate, data_labels=None, title_correspondance=None, for_frames=False):
        """
        Compare with a wilcoxon test the performance over all pair of labels (methods)
        Args:
            stats_to_evaluate: list of str
            title_correspondance: dict

        Returns:

        """
        print("")
        print("-" * 100)
        print("COMPARING METRICS values FOR EACH LABEL (Wilcoxon signed-rank test)")
        result_dict_to_use = self.results_frames_dict
        if not for_frames:
            result_dict_to_use = self.results_transients_dict

        for stat_name in stats_to_evaluate:
            stat_name_to_display = stat_name
            if (title_correspondance is not None) and (stat_name in title_correspondance):
                stat_name_to_display = title_correspondance[stat_name]
            print(f"## For {stat_name_to_display}:")
            n_labels = len(data_labels)
            if data_labels is None:
                data_labels = list(result_dict_to_use.keys())

            values_by_prediction = [[] for n in np.arange(n_labels)]
            cells_by_prediction = [[] for n in np.arange(n_labels)]

            for label_index, label in enumerate(data_labels):
                session_dict = result_dict_to_use[label]
                for session_id, cells_dict in session_dict.items():
                    for cell_to_display, stat_dict in cells_dict.items():
                        if stat_name == "F1":
                            recall = stat_dict["sensitivity"]
                            precision = stat_dict["PPV"]
                            if (precision + recall) == 0:
                                f1_score = 0
                            else:
                                f1_score = 2 * ((precision * recall) / (precision + recall))
                                values_by_prediction[label_index].append(f1_score)
                        else:
                            # putting metric value for each cell in the list
                            values_by_prediction[label_index].append(stat_dict[stat_name])
                        cells_by_prediction[label_index].append(session_id + "_" + str(cell_to_display))

            for label_index in np.arange(n_labels - 1):
                for label_index_2 in np.arange(1, n_labels):
                    if label_index == label_index_2:
                        continue
                    values = np.array(values_by_prediction[label_index])
                    values_2 = np.array(values_by_prediction[label_index_2])
                    common_cells = np.intersect1d(cells_by_prediction[label_index],
                                                  cells_by_prediction[label_index_2])
                    # common_cells = list(set(cells_by_prediction[label_index]) &
                    #                     set(cells_by_prediction[label_index_2]))
                    # same cells
                    if (len(values) != len(values_2)) or len(values) != len(common_cells):
                        # we only keep common cells
                        # list of string
                        cells = cells_by_prediction[label_index]
                        cells_2 = cells_by_prediction[label_index_2]

                        values = [values[cells.index(cell)] for cell in common_cells]
                        values_2 = [values_2[cells_2.index(cell)] for cell in common_cells]
                    sum_of_ranks, p_value = wilcoxon(values, values_2)

                    print(f"{data_labels[label_index]} vs {data_labels[label_index_2]}: p = "
                          f"{np.round(p_value, 3)}, T = {np.round(sum_of_ranks, 1)}")

            print("")

        print("-" * 100)
        print("")

    def plot_boxplots_f1_score(self, description, colorfull=False, time_str=None,
                               for_frames=True, with_cells=False,
                               color_cell_as_boxplot=False,
                               box_plots_labels=None,
                               white_background=False,
                               with_legend=False,
                               using_patch_for_legend=True,
                               alpha_scatter=1, with_cell_number=True,
                               put_metric_as_y_axis_label=False,
                               save_formats="pdf", dpi=500):
        """

        :param description:
        :param time_str:
        :param colorfull (bool): if True, the boxplots are filled with color, one different for each
        (based on brewer colors)
        :param white_background: (bool) if True white background, else black background
        :param for_frames:
        :param with_cells: if True, display a scatter for each cell
        :param save_formats:
        :return:
        """
        result_dict_to_use = self.results_frames_dict
        if not for_frames:
            result_dict_to_use = self.results_transients_dict

        stat_fig, ax = plt.subplots(nrows=1, ncols=1, squeeze=True,
                                    gridspec_kw={'height_ratios': [1],
                                                 'width_ratios': [1]},
                                    figsize=(5, 5), dpi=dpi)

        stat_fig.set_tight_layout({'rect': [0, 0, 1, 1], 'pad': 1, 'w_pad': 1, 'h_pad': 5})
        fig_patch = stat_fig.patch

        # rgba = c_map(0)
        if white_background:
            face_color = "white"
            text_color = "black"
            title_color = "black"
        else:
            face_color = "black"
            text_color = "white"
            title_color = "red"
        fig_patch.set_facecolor(face_color)

        # n_cells = len(self.results_frames_dict_by_cell)

        ax.set_facecolor(face_color)

        ax.set_frame_on(False)

        n_box_plots = len(result_dict_to_use)
        if box_plots_labels is None:
            labels = list(result_dict_to_use.keys())
        else:
            labels = box_plots_labels
        values_by_prediction = [[] for n in np.arange(n_box_plots)]

        for label_index, label in enumerate(labels):
            session_dict = result_dict_to_use[label]
            for session_id, cells_dict in session_dict.items():
                for cell_to_display, stat_dict in cells_dict.items():
                    recall = stat_dict["sensitivity"]
                    precision = stat_dict["PPV"]
                    if (precision + recall) == 0:
                        f1_score = 0
                    else:
                        f1_score = 2 * ((precision * recall) / (precision + recall))
                    values_by_prediction[label_index]. \
                        append(f1_score)
                    if with_cells:
                        # Adding jitter
                        x_pos = 1 + label_index + ((np.random.random_sample() - 0.5) * 0.7)
                        y_pos = f1_score
                        font_size = 3
                        if white_background:
                            edgecolors = "black"
                        else:
                            edgecolors = "white"

                        if color_cell_as_boxplot:
                            color_cell = self.colors_boxplots[label_index % len(self.colors_boxplots)]
                        elif session_id not in self.cells_to_color:
                            color_cell = BREWER_COLORS[label_index % len(BREWER_COLORS)]
                        else:
                            color_cell = self.cells_to_color[session_id]
                        if label_index == 0 and with_legend:
                            ax.scatter(x_pos, y_pos,
                                       color=color_cell,
                                       marker="o",
                                       label=session_id,
                                       edgecolors=edgecolors,
                                       s=60, zorder=21, alpha=alpha_scatter)
                        else:
                            ax.scatter(x_pos, y_pos,
                                       color=color_cell,
                                       marker="o",
                                       edgecolors=edgecolors,
                                       s=60, zorder=21, alpha=alpha_scatter)
                        if with_cell_number:
                            ax.text(x=x_pos, y=y_pos,
                                    s=f"{cell_to_display}", color="black", zorder=22,
                                    ha='center', va="center", fontsize=font_size, fontweight='bold')

        outliers = dict(markerfacecolor='white', marker='D')
        print(f"F1 SCORE for_frames {for_frames}: {description}")
        for label_index, label in enumerate(labels):
            print(f"{label}: median: {np.round(np.median(values_by_prediction[label_index]), 3)}, "
                  f"25p {np.round(np.percentile(values_by_prediction[label_index], 25), 3)}, "
                  f"75p {np.round(np.percentile(values_by_prediction[label_index], 75), 3)}")
        bplot = ax.boxplot(values_by_prediction, patch_artist=colorfull,
                           flierprops=outliers, widths=[0.7] * len(values_by_prediction),
                           labels=labels, sym='', zorder=1)  # whis=[5, 95], sym='+'

        for element in ['boxes', 'whiskers', 'fliers', 'caps']:
            if white_background:
                plt.setp(bplot[element], color="black")
            else:
                plt.setp(bplot[element], color="white")

        for element in ['means', 'medians']:
            if white_background:
                plt.setp(bplot[element], color="black")
            else:
                plt.setp(bplot[element], color="white")  # "silver"

        if colorfull:
            colors = self.colors_boxplots[:n_box_plots]
            for patch, color in zip(bplot['boxes'], colors):
                patch.set_facecolor(color)

        # adding legend if cells displayed
        if with_cells and with_legend and (not color_cell_as_boxplot):
            if using_patch_for_legend:
                legend_elements = []
                # looking for all session_ids
                all_session_ids = set()
                for label_index, label in enumerate(labels):
                    session_dict = result_dict_to_use[label]
                    for session_id, cells_dict in session_dict.items():
                        all_session_ids.add(session_id)
                all_session_ids = list(all_session_ids)
                all_session_ids.sort()
                for session_id in all_session_ids:
                    if session_id not in self.cells_to_color:
                        continue
                    color_cell = self.cells_to_color[session_id]

                    legend_elements.append(Patch(facecolor=color_cell, hatch="o",
                                                 edgecolor='black', label=session_id))

                # if use_different_shapes_for_stat:
                #     for cat in np.arange(1, n_categories + 1):
                #         if cat in banned_categories:
                #             continue
                #         legend_elements.append(Line2D([0], [0], marker=param.markers[cat - 1], color="w", lw=0, label="*" * cat,
                #                                       markerfacecolor='black', markersize=15))
                #
                ax.legend(handles=legend_elements)
            else:
                # using the label put in the scatter
                ax.legend()

        ax.xaxis.set_ticks_position('none')
        ax.xaxis.label.set_color(text_color)
        ax.tick_params(axis='x', colors=text_color)

        if n_box_plots <= 2:
            ax.xaxis.set_tick_params(labelsize=20)
        elif n_box_plots <= 4:
            ax.xaxis.set_tick_params(labelsize=15)
        else:
            ax.xaxis.set_tick_params(labelsize=2)
        ax.yaxis.label.set_color(text_color)
        ax.tick_params(axis='y', colors=text_color)
        # ax.set_xticklabels([])
        # ax.set_yticklabels([])
        # ax.get_yaxis().set_visible(False)
        # ax.get_xaxis().set_visible(False)
        # ax.set_ylabel(f"proportion")
        # ax.set_xlabel("age")
        xticks = np.arange(1, n_box_plots + 1)
        ax.set_xticks(xticks)
        # sce clusters labels
        ax.set_xticklabels(labels)
        yticks = np.arange(0, 1.2, 0.2)
        ax.set_yticks(yticks)
        ax.set_yticklabels(np.arange(0, 120, 20))
        # fixing the limits
        ax.set_ylim(0, 1.1)

        if put_metric_as_y_axis_label:
            ax.set_ylabel("F1 SCORE" " (%)")
            ax.yaxis.label.set_size(20)
        else:
            ax.set_title("F1 SCORE", color=title_color, pad=20, fontsize=20)

        str_details = "frames"
        if not for_frames:
            str_details = "transients"

        if time_str is None:
            time_str = datetime.now().strftime("%Y_%m_%d.%H-%M-%S")

        if isinstance(save_formats, str):
            save_formats = [save_formats]
        for save_format in save_formats:
            stat_fig.savefig(f'{self.results_path}/'
                             f'{description}_box_plots_f1_score_{str_details}'
                             f'_{time_str}.{save_format}',
                             format=f"{save_format}",
                             facecolor=stat_fig.get_facecolor(), edgecolor='none')
        plt.close()

    def plot_boxplots_proportion_frames_in_transients(self, description, time_str=None,
                                                      colorfull=True,
                                                      white_background=False,
                                                      only_this_key=None,
                                                      with_scatter=True,
                                                      with_cell_text=True,
                                                      save_formats="pdf", dpi=500):
        """

        :param description:
        :param time_str:
        param only_this_key: (str): key of the label, if not None, then only the boxplot corresponding will be displayed
        :param save_formats:
        :return:
        """
        result_dict_to_use = self.results_frames_in_transients_pc

        stat_fig, ax = plt.subplots(nrows=1, ncols=1, squeeze=True,
                                    gridspec_kw={'height_ratios': [1],
                                                 'width_ratios': [1]},
                                    figsize=(5, 10), dpi=dpi)

        stat_fig.set_tight_layout({'rect': [0, 0, 1, 1], 'pad': 1, 'w_pad': 1, 'h_pad': 5})
        fig_patch = stat_fig.patch

        # rgba = c_map(0)
        if white_background:
            face_color = "white"
            text_color = "black"
            title_color = "black"
        else:
            face_color = "black"
            text_color = "white"
            title_color = "red"
        fig_patch.set_facecolor(face_color)

        # n_cells = len(self.results_frames_dict_by_cell)

        ax.set_facecolor(face_color)

        ax.set_frame_on(False)

        if only_this_key and only_this_key in result_dict_to_use:
            n_box_plots = 1
            labels = [only_this_key]
        else:
            n_box_plots = len(result_dict_to_use)
            labels = list(result_dict_to_use.keys())

        values_by_prediction = [[] for n in np.arange(n_box_plots)]

        for label_index, label in enumerate(labels):
            session_dict = result_dict_to_use[label]
            for session_id, cells_dict in session_dict.items():
                for cell_to_display, pourcentages in cells_dict.items():
                    values_by_prediction[label_index].extend(pourcentages)

                    if with_scatter:
                        for pc in pourcentages:
                            # Adding jitter
                            x_pos = 1 + label_index + ((np.random.random_sample() - 0.5) * 0.7)
                            y_pos = pc
                            font_size = 3
                            if white_background:
                                edgecolors = "black"
                            else:
                                edgecolors = "white"
                            if session_id not in self.cells_to_color:
                                color_cell = BREWER_COLORS[label_index % len(BREWER_COLORS)]
                            else:
                                color_cell = self.cells_to_color[session_id]
                            ax.scatter(x_pos, y_pos,
                                       color=color_cell,
                                       marker="o",
                                       edgecolors=edgecolors,
                                       s=60, zorder=21, alpha=0.4)
                            if cell_to_display > 999:
                                font_size = 2
                            if with_cell_text:
                                ax.text(x=x_pos, y=y_pos,
                                        s=f"{cell_to_display}", color="black", zorder=22,
                                        ha='center', va="center", fontsize=font_size, fontweight='bold')

        outliers = dict(markerfacecolor='white', marker='D')
        print("proportion_frames_in_transients")
        for label_index, label in enumerate(labels):
            print(f"{label}: median: {np.round(np.median(values_by_prediction[label_index]), 3)}, "
                  f"25p {np.round(np.percentile(values_by_prediction[label_index], 25), 3)}, "
                  f"75p {np.round(np.percentile(values_by_prediction[label_index], 75), 3)}")
        bplot = ax.boxplot(values_by_prediction, patch_artist=colorfull,
                           flierprops=outliers, widths=[0.7] * len(values_by_prediction),
                           labels=labels, sym='', zorder=1)  # whis=[5, 95], sym='+'

        for element in ['boxes', 'whiskers', 'fliers', 'caps']:
            if white_background:
                plt.setp(bplot[element], color="black")
            else:
                plt.setp(bplot[element], color="white")

        for element in ['means', 'medians']:
            if white_background:
                plt.setp(bplot[element], color="black")
            else:
                plt.setp(bplot[element], color="white")  # silver

        if colorfull:
            colors = self.colors_boxplots[:n_box_plots]
            for patch, color in zip(bplot['boxes'], colors):
                patch.set_facecolor(color)

        ax.xaxis.set_ticks_position('none')
        ax.xaxis.label.set_color(text_color)
        ax.tick_params(axis='x', colors=text_color)
        if n_box_plots <= 2:
            ax.xaxis.set_tick_params(labelsize=15)
        elif n_box_plots <= 6:
            ax.xaxis.set_tick_params(labelsize=9)
        else:
            ax.xaxis.set_tick_params(labelsize=2)
        ax.yaxis.label.set_color(text_color)
        ax.tick_params(axis='y', colors=text_color)
        # ax.set_xticklabels([])
        # ax.set_yticklabels([])
        # ax.get_yaxis().set_visible(False)
        # ax.get_xaxis().set_visible(False)
        # ax.set_ylabel(f"proportion")
        # ax.set_xlabel("age")
        xticks = np.arange(1, n_box_plots + 1)
        ax.set_xticks(xticks)
        # sce clusters labels
        ax.set_xticklabels(labels)
        # fixing the limits
        # ax.set_ylim(0, 100)

        # ax.set_title("", color=title_color, pad=20, fontsize=20)

        str_details = ""
        if only_this_key:
            str_details = "_" + only_this_key
        if time_str is None:
            time_str = datetime.now().strftime("%Y_%m_%d.%H-%M-%S")

        if isinstance(save_formats, str):
            save_formats = [save_formats]
        for save_format in save_formats:
            stat_fig.savefig(f'{self.results_path}/'
                             f'{description}_box_plots_proportion_frames_in_transients{str_details}'
                             f'_{time_str}.{save_format}',
                             format=f"{save_format}",
                             facecolor=stat_fig.get_facecolor(), edgecolor='none')
        plt.close()


def load_data_from_np_or_mat_file(file_name, attr_name=None):
    """
    Load data from a numpy or matlab file (.npz, .npz or .mat)
    Args:
        file_name:
        data_descr: string used to display error message if the file is not in the good format
        attr_name:

    Returns:


    """
    if file_name.endswith(".npy"):
        data = np.load(file_name)
    elif file_name.endswith(".npz"):
        data = np.load(file_name)
        if attr_name not in data:
            raise Exception(f"{attr_name} not available in {file_name}, choices are {list(data.keys())}")
        data = data[attr_name]
    elif file_name.endswith(".mat"):  # .mat
        data = hdf5storage.loadmat(file_name)
        if attr_name not in data:
            raise Exception(f"{attr_name} not available in {file_name}, choices are {list(data.keys())}")
        data = data[attr_name]
    else:
        print("Load_data_from_np_or_mat_file() error: file extension must be .npy, .npz or .mat")
        return

    return data


def bin_raster(spike_nums):
    """
    Take a binary 2d array (n_cell*n_frames) and return a binned version,
    with output shape 2d array (n_cell*n_frames//2)
    Args:
        spike_nums:

    Returns:

    """
    spike_nums_bin = np.zeros((spike_nums.shape[0], spike_nums.shape[1] // 2),
                              dtype="int8")
    for cell in np.arange(spike_nums_bin.shape[0]):
        binned_cell = spike_nums[cell].reshape(-1, 2).mean(axis=1)
        binned_cell[binned_cell > 0] = 1
        spike_nums_bin[cell] = binned_cell.astype("int")

    return spike_nums_bin


def get_raster_dur_spikes_and_traces(spike_nums, traces):
    """
    Take a binary raster (n_cells*n_frames) representing spikes or onsets of transient,
    and based on the smoothed traces, detect putative transients and extend the spikes/onsets
    so that the duration of the rise time of the transient in which they are fully 'active'
    Args:
        spike_nums:
        traces:

    Returns:

    """

    # to work just on onsets/spikes
    # return caiman_spike_nums

    n_cells = traces.shape[0]
    n_times = traces.shape[1]

    for i in np.arange(n_cells):
        traces[i, :] = (traces[i, :] - np.mean(traces[i, :])) / np.std(traces[i, :])

    spike_nums_all = np.zeros((n_cells, n_times), dtype="int8")
    for cell in np.arange(n_cells):
        onsets = []
        diff_values = np.diff(traces[cell])
        for index, value in enumerate(diff_values):
            if index == (len(diff_values) - 1):
                continue
            if value < 0:
                if diff_values[index + 1] >= 0:
                    onsets.append(index + 1)
        if len(onsets) > 0:
            spike_nums_all[cell, np.array(onsets)] = 1

    peak_nums = np.zeros((n_cells, n_times), dtype="int8")
    for cell in np.arange(n_cells):
        peaks, properties = signal.find_peaks(x=traces[cell])
        peak_nums[cell, peaks] = 1

    spike_nums_dur = build_spike_nums_dur(spike_nums_all, peak_nums)

    final_spike_nums_dur = np.zeros((spike_nums_dur.shape[0], spike_nums_dur.shape[1]), dtype="int8")
    for cell in np.arange(n_cells):
        periods = get_continous_time_periods(spike_nums_dur[cell])
        for period in periods:
            if np.sum(spike_nums[cell, period[0]:period[1] + 1]) > 0:
                final_spike_nums_dur[cell, period[0]:period[1] + 1] = 1

    return final_spike_nums_dur


class SessionForBenchmark:
    """
    From a directory content including a yaml config file, will read ground truth + infered data
    """

    def __init__(self, dir_to_explore, inferences_to_benchmark, predictions_keywords_dict,
                 default_predictions_threshold):
        """

        Args:
            dir_to_explore:
            inferences_to_benchmark: list of inference_id, if empty with evaluate all inference, otherwise just those in the list
        """
        self.inferences_to_benchmark = inferences_to_benchmark
        self.dir_to_explore = dir_to_explore
        self.valid = False
        self.session_id = os.path.basename(dir_to_explore)
        self.predictions_keywords_dict = predictions_keywords_dict
        self.default_predictions_threshold = default_predictions_threshold
        self.cells_in_ground_truth = None

        # 2d binary array (n_cells * n_frames)
        self.ground_truth = None
        self.smooth_traces = None
        self.inferences_raster_dict = dict()
        self.cells_by_inference = dict()

        file_names = []

        # look for filenames in the fisrst directory, if we don't break, it will go through all directories
        for (dirpath, dirnames, local_filenames) in os.walk(self.dir_to_explore):
            file_names.extend(local_filenames)
            break

        yaml_file_name = [f for f in file_names if f.endswith(".yaml")]
        if len(yaml_file_name) == 0:
            print(f"SessionForBenchmark: no yaml config file in : {self.dir_to_explore}")
            return
        yaml_file_name = os.path.join(dir_to_explore, yaml_file_name[0])

        file_names = [f for f in file_names if (not f.endswith(".yaml")) and (not f.startswith("."))]

        # reading the yaml file
        with open(yaml_file_name, 'r') as stream:
            yaml_data = yaml.load(stream, Loader=yaml.FullLoader)

        if "groundtruth" not in yaml_data:
            raise Exception(f"No groundtruth given in {yaml_file_name}")

        yaml_keys = list(yaml_data.keys())
        # we want to start by ground_truth
        yaml_keys.remove("groundtruth")
        yaml_keys = ["groundtruth"] + yaml_keys
        for inference_id in yaml_keys:
            inference_data = yaml_data[inference_id]
            if "cells" not in inference_data:
                raise Exception(f"Cells are not indicated for {inference_id} of {yaml_file_name}")
            cells = inference_data["cells"]

            if "file_name" not in inference_data:
                raise Exception(f"No file_name given in inference data {inference_id} of {yaml_file_name}")
            inference_file_name = inference_data["file_name"]

            # we remove the file from the list of filenames
            if inference_file_name in file_names:
                file_names.remove(inference_file_name)

            inference_file_name = os.path.join(dir_to_explore, inference_file_name)

            if inference_id == "groundtruth":
                # then we look at extension to decide how to benchmark it
                if not inference_file_name.endswith("cinac"):
                    raise Exception(f"{os.path.basename(inference_file_name)} format not yet "
                                    f"implemented for ground truth"
                                    f"in {yaml_file_name}")
                self.cells_in_ground_truth = cells
                # so far ground truth file has to be in cinac format
                self.extract_data_from_cinac_file(cinac_file=inference_file_name, cells_to_add=cells,
                                                  for_ground_truth=True, label=None)
                continue
            else:
                if len(self.inferences_to_benchmark) > 0 and (inference_id not in self.inferences_to_benchmark):
                    # this inference is not on the list of inferences to evaluate
                    continue

                if inference_file_name.endswith("cinac"):
                    self.extract_data_from_cinac_file(cinac_file=inference_file_name, cells_to_add=cells,
                                                      for_ground_truth=False, label=inference_id)
                    continue
                if "prediction_threshold" in inference_data:
                    # means the data is issue of predictions from cinac
                    prediction_threshold = inference_data["prediction_threshold"]
                    predictions = load_data_from_np_or_mat_file(file_name=inference_file_name,
                                                                attr_name="predictions")

                    # then we produce the raster dur based on the predictions using threshold the prediction_threshold
                    raster_dur = np.zeros(predictions.shape, dtype="int8")

                    for cell in np.arange(len(predictions)):
                        raster_dur[cell, predictions[cell] >= prediction_threshold] = 1
                    self.inferences_raster_dict[inference_id] = [raster_dur, predictions]
                    self.cells_by_inference[inference_id] = cells
                    continue
                if inference_file_name.endswith("npy") or inference_file_name.endswith("mat") or \
                        inference_file_name.endswith("npz"):
                    # attribute name in the file (for npz or mat files)
                    attr_name = inference_data.get("attr_name", None)
                    # get a 2d array of shape n_cells*n_frames
                    # print(f"inference_file_name {inference_file_name}")
                    inference_neuronal_data = load_data_from_np_or_mat_file(file_name=inference_file_name,
                                                                            attr_name=attr_name)
                    if "to_bin" in inference_data and inference_data['to_bin']:
                        # bining the data, useful for caiman output for exemple
                        inference_neuronal_data = bin_raster(inference_neuronal_data)
                    if "traces_file_name" in inference_data:
                        traces_file_name = inference_data["traces_file_name"]
                        traces_file_name = os.path.join(dir_to_explore, traces_file_name)
                        traces_attr_name = inference_data.get("traces_attr_name", None)
                        traces = load_data_from_np_or_mat_file(file_name=traces_file_name,
                                                               attr_name=traces_attr_name)
                        inference_neuronal_data = get_raster_dur_spikes_and_traces(spike_nums=inference_neuronal_data,
                                                                                   traces=traces)

                    # mapping fiji coord to caiman coord if necessary
                    if "caiman_fiji_mapping" in inference_data:
                        print("caiman_fiji_mapping in inference_data")
                        caiman_fiji_mapping = np.load(
                            os.path.join(dir_to_explore, inference_data["caiman_fiji_mapping"]))
                        caiman_raster_dur = np.zeros((self.ground_truth.shape[0], inference_neuronal_data.shape[1]),
                                                     dtype="int8")
                        cells_mapped = []
                        # print(f"caiman_fiji_mapping {caiman_fiji_mapping}")
                        # caiman_fiji_mapping take a cell index from caiman segmentation
                        # and return a value > 0 if a cell of Fiji matches (the value is the index of this cell in fiji)
                        # -1 otherwise
                        for caiman_cell_index, fiji_cell_index in enumerate(caiman_fiji_mapping):
                            if fiji_cell_index >= 0:
                                # print(f"{inference_id}: Cell {fiji_cell_index} -> {caiman_cell_index}")
                                # using deconvolution value, cell is active if value > 0
                                caiman_raster_dur[fiji_cell_index] = inference_neuronal_data[caiman_cell_index]
                                # print(f"sum spikes: {np.sum(caiman_raster_dur[fiji_cell_index])}")
                                cells_mapped.append(fiji_cell_index)
                        inference_neuronal_data = caiman_raster_dur
                        cells_to_keep = []
                        for cell in cells:
                            if cell not in cells_mapped:  # caiman_fiji_mapping:
                                print(f"Cell {cell} has no match in caiman segmentation")
                            else:
                                cells_to_keep.append(cell)
                        cells = np.array(cells_to_keep)
                    self.inferences_raster_dict[inference_id] = [inference_neuronal_data]
                    self.cells_by_inference[inference_id] = cells
                    continue
                else:
                    print(f"Format of {os.path.basename(inference_file_name)} not supported")

        # then if the list of prediction keywords is not None, we list the files to find some
        if len(predictions_keywords_dict) > 0:
            for pred_keyword, inference_id in predictions_keywords_dict.items():
                for file_name in file_names:
                    if pred_keyword in file_name:
                        inference_file_name = os.path.join(dir_to_explore, file_name)
                        prediction_threshold = default_predictions_threshold
                        predictions = load_data_from_np_or_mat_file(file_name=inference_file_name,
                                                                    attr_name="predictions")

                        # then we produce the raster dur based on the predictions using threshold the prediction_threshold
                        raster_dur = np.zeros(predictions.shape, dtype="int8")

                        for cell in np.arange(len(predictions)):
                            raster_dur[cell, predictions[cell] >= prediction_threshold] = 1
                        self.inferences_raster_dict[inference_id] = [raster_dur, predictions]
                        # we put the same cells as in GT
                        self.cells_by_inference[inference_id] = self.cells_in_ground_truth
                        break

        self.valid = True

    def get_inference_ids(self):
        return list(self.inferences_raster_dict.keys())

    def extract_data_from_cinac_file(self, cinac_file, cells_to_add, for_ground_truth, label=None):
        """
        From a cinac_file,
        Args:
            session_id: str
            cinac_file:
            cells_to_add:
            for_ground_truth: (bool) to set ground truth or add some inference
            label: (str), necessary if for_ground_truth is False, allow to identify the inference

        Returns:

        """

        cinac_file_reader = CinacFileReader(file_name=cinac_file)

        n_cells = cinac_file_reader.get_n_cells()

        n_frames = cinac_file_reader.get_n_frames()

        if n_cells is None:
            raise Exception(f"{os.path.basename(cinac_file)} in {self.session_id} has not the n_cells info")

        if n_frames is None:
            raise Exception(f"{os.path.basename(cinac_file)} in {self.session_id} has not the n_frames info")

        segments = cinac_file_reader.get_all_segments()

        raster_dur = np.zeros((n_cells, n_frames), dtype="int8")

        if for_ground_truth:
            self.smooth_traces = np.zeros((n_cells, n_frames))

        for segment in segments:
            cell = segment[0]
            first_frame = segment[1]
            last_frame = segment[2]
            if cell not in cells_to_add:
                continue
            # then we fill the raster_dur
            raster_dur[cell, first_frame:last_frame + 1] = cinac_file_reader.get_segment_raster_dur(segment=segment)
            # and fluoresence signal, needed for the ground truth in order to find the putative transients
            if for_ground_truth:
                self.smooth_traces[cell, first_frame:last_frame + 1] = cinac_file_reader. \
                    get_segment_smooth_traces(segment=segment)

        if for_ground_truth:
            self.ground_truth = raster_dur
        else:
            self.inferences_raster_dict[label] = [raster_dur]
            self.cells_by_inference[label] = cells_to_add

    def add_it_to_cinac_benchmark(self, cinac_benchmarks):
        cinac_benchmarks.add_ground_truth(session_id=self.session_id, ground_truth=self.ground_truth,
                                          smooth_traces=self.smooth_traces)

        for inference_id, inference_data in self.inferences_raster_dict.items():
            raster_predictions = None
            if len(inference_data) > 1:
                raster_predictions = inference_data[1]
            cinac_benchmarks.add_inference_to_benchmark(session_id=self.session_id,
                                                        inference_to_benchmark_id=inference_id,
                                                        raster_to_evaluate=inference_data[0],
                                                        cells_to_benchmark=self.cells_by_inference[inference_id],
                                                        raster_predictions=raster_predictions)


def do_traces_smoothing(traces):
    # smoothing the trace
    windows = ['hanning', 'hamming', 'bartlett', 'blackman']
    i_w = 1
    window_length = 7  # 11
    for i in np.arange(traces.shape[0]):
        smooth_signal = smooth_convolve(x=traces[i], window_len=window_length,
                                        window=windows[i_w])
        beg = (window_length - 1) // 2
        traces[i] = smooth_signal[beg:-beg]


def benchmark_neuronal_activity_inferences(inferences_dir, results_path, colorfull_boxplots=True,
                                           white_background=False, color_cell_as_boxplot=False,
                                           with_legend=False, put_metric_as_y_axis_label=False,
                                           using_patch_for_legend=True,alpha_scatter=0.6,
                                           with_cells=True, with_cell_number=True,
                                           predictions_stat_by_metrics=False):
    """
    Evaluate the performances of different inferences methods. Every directory in the main
    directory is considered a session with a ground truth and inferences to benchmark
    Args:
        inferences_dir: directory containing other directories, one by session to benchmark
        results_path:
        colorfull_boxplots: if True boxplots are filled with color
        white_background: if True background is white, else it's black
        with_cells: if True display scatter to represent metrics for a cell
        with_cell_number: Displau the cell number in the scatter if with_cells is True
        with_legend: if True, add legens to the F1 score plot

    Returns:

    """

    directories = []
    file_names_in_root = []

    # look for filenames in the fisrst directory, if we don't break, it will go through all directories
    for (dirpath, dirnames, local_filenames) in os.walk(inferences_dir):
        directories.extend(dirnames)
        file_names_in_root.extend(local_filenames)
        break
    if len(directories) == 0:
        print(f"No directories with inferences in {inferences_dir}")
        return

    config_yaml_files = [f for f in file_names_in_root if (not f.startswith(".") and f.endswith(".yaml"))]
    if len(config_yaml_files) == 0:
        print(f"No yaml benchmarks config file in {inferences_dir}")
        return

    # looping over configuration files and evaluate inference with each config
    for config_yaml_file in config_yaml_files:
        sessions_for_benchmark = []
        # first getting the main config yaml
        # list of inference_id, if empty with evaluate all inference, otherwise just those in the list
        inferences_to_benchmark = []
        sessions_to_exclude = []
        # dict that contains as keys a list of keyword allowing to selecting predictions file automatically
        # the value is the inference_id
        predictions_keywords_dict = {}
        default_predictions_threshold = 0.5

        config_yaml_file = os.path.join(inferences_dir, config_yaml_file)
        with open(config_yaml_file, 'r') as stream:
            yaml_data = yaml.load(stream, Loader=yaml.FullLoader)
        inferences_to_benchmark = yaml_data.get("inferences_to_benchmark", [])
        sessions_to_exclude = yaml_data.get("sessions_to_exclude", [])
        predictions_keywords_dict = yaml_data.get("predictions_keywords", {})
        default_predictions_threshold = yaml_data.get("predictions_threshold", 0.5)
        # allows to create a subdirecoty to put the results
        config_id = yaml_data.get("config_id", None)

        if config_id is None:
            config_results_path = results_path
        else:
            # create a directory
            config_results_path = os.path.join(results_path, config_id)
            os.mkdir(config_results_path)

        directories = [os.path.join(inferences_dir, d) for d in directories if not d.startswith(".")]

        inference_ids = []

        for dir_to_explore in directories:
            if os.path.basename(dir_to_explore) in sessions_to_exclude:
                # excluding the session evaluation
                continue

            session_for_benchmark = SessionForBenchmark(dir_to_explore=dir_to_explore,
                                                        inferences_to_benchmark=inferences_to_benchmark,
                                                        predictions_keywords_dict=predictions_keywords_dict,
                                                        default_predictions_threshold=default_predictions_threshold)
            if session_for_benchmark.valid:
                sessions_for_benchmark.append(session_for_benchmark)
                inference_ids.extend(session_for_benchmark.get_inference_ids())

        # making it unique
        inference_ids = set(inference_ids)
        colors_boxplots = ['#a6cee3', '#1f78b4', '#b2df8a', '#33a02c', '#fb9a99', '#e31a1c',
                           '#ff7f00', '#cab2d6', '#fdbf6f', '#6a3d9a', '#ffff99', '#b15928', '#ffffd9']
        cinac_benchmarks = CinacBenchmarks(results_path=config_results_path, colors_boxplots=colors_boxplots)

        description = ""
        for session_index, session_for_benchmark in enumerate(sessions_for_benchmark):
            description = description + session_for_benchmark.session_id + "_"
            session_for_benchmark.add_it_to_cinac_benchmark(cinac_benchmarks=cinac_benchmarks)
            cinac_benchmarks.color_by_session(session_id=session_for_benchmark.session_id,
                                              color=BREWER_COLORS[session_index % len(BREWER_COLORS)])

        cinac_benchmarks.evaluate_metrics()

        time_str = datetime.now().strftime("%Y_%m_%d.%H-%M-%S")

        # only transients
        for for_frames in [False]:
            if predictions_stat_by_metrics:
                # plot the distribution of the the prediction values for each cell
                cinac_benchmarks.plot_boxplot_predictions_stat_by_metrics(description=description, time_str=time_str,
                                                                          for_frames=for_frames,
                                                                          colorfull=colorfull_boxplots,
                                                                          white_background=white_background,
                                                                          save_formats=["png"],
                                                                          dpi=300)

            # tp get PPV and Sensitivity in the same figure
            # cinac_benchmarks.plot_boxplots_full_stat(description=description, time_str=time_str,
            #                                          colorfull=colorfull_boxplots,
            #                                          color_cell_as_boxplot=color_cell_as_boxplot,
            #                                          box_plots_labels=inferences_to_benchmark,
            #                                          white_background=white_background,
            #                                          stats_to_show=("sensitivity", "PPV"),
            #                                          title_correspondance={"sensitivity": "Sensitivity",
            #                                                                "PPV": "Precision"},
            #                                          for_frames=for_frames, save_formats=["png"], with_cells=True,
            #                                          alpha_scatter=0.8,
            #                                          dpi=500)

            cinac_benchmarks.stats_on_performance(stats_to_evaluate=["sensitivity", "PPV", "F1"],
                                                  data_labels=inferences_to_benchmark,
                                                  title_correspondance={"PPV": "Precision",
                                                                        "sensitivity": "Sensitivity"})

            cinac_benchmarks.plot_boxplots_full_stat(description=description, time_str=time_str,
                                                     colorfull=colorfull_boxplots,
                                                     color_cell_as_boxplot=color_cell_as_boxplot,
                                                     box_plots_labels=inferences_to_benchmark,
                                                     white_background=white_background,
                                                     stats_to_show=["sensitivity"],
                                                     title_correspondance={"sensitivity": "Sensitivity"},
                                                     for_frames=for_frames, save_formats=["png"], with_cells=with_cells,
                                                     put_metric_as_y_axis_label=put_metric_as_y_axis_label,
                                                     with_cell_number=with_cell_number,
                                                     alpha_scatter=alpha_scatter,
                                                     dpi=500)

            cinac_benchmarks.plot_boxplots_full_stat(description=description, time_str=time_str,
                                                     colorfull=colorfull_boxplots,
                                                     color_cell_as_boxplot=color_cell_as_boxplot,
                                                     box_plots_labels=inferences_to_benchmark,
                                                     white_background=white_background,
                                                     stats_to_show=["PPV"],
                                                     title_correspondance={"PPV": "Precision"},
                                                     for_frames=for_frames, save_formats=["png"], with_cells=with_cells,
                                                     put_metric_as_y_axis_label=put_metric_as_y_axis_label,
                                                     with_cell_number=with_cell_number,
                                                     alpha_scatter=alpha_scatter,
                                                     dpi=500)

            cinac_benchmarks.plot_boxplots_f1_score(description=description, time_str=time_str,
                                                    colorfull=colorfull_boxplots,
                                                    color_cell_as_boxplot=color_cell_as_boxplot,
                                                    box_plots_labels=inferences_to_benchmark,
                                                    white_background=white_background,
                                                    with_cell_number=with_cell_number,
                                                    with_legend=with_legend,
                                                    alpha_scatter=alpha_scatter,
                                                    using_patch_for_legend=using_patch_for_legend,
                                                    for_frames=for_frames, save_formats=["png"], with_cells=with_cells,
                                                    put_metric_as_y_axis_label=put_metric_as_y_axis_label,
                                                    dpi=500)

        # cinac_benchmarks.plot_boxplots_proportion_frames_in_transients(description=description,
        #                                                                time_str=time_str,
        #                                                                only_this_key=None,
        #                                                                colorfull=colorfull_boxplots,
        #                                                                white_background=white_background,
        #                                                                with_scatter=True,
        #                                                                with_cell_text=False,
        #                                                                save_formats="png", dpi=500)
        #
        # for inference_id in inference_ids:
        #     cinac_benchmarks.plot_boxplots_proportion_frames_in_transients(description=description,
        #                                                                    time_str=time_str,
        #                                                                    only_this_key=inference_id,
        #                                                                    colorfull=colorfull_boxplots,
        #                                                                    white_background=white_background,
        #                                                                    with_scatter=True,
        #                                                                    with_cell_text=False,
        #                                                                    save_formats="png", dpi=500)



import tensorflow as tf

TF_VERSION = tf.__version__
print(f"TF_VERSION {TF_VERSION}")

# depending on the TF version installed
if TF_VERSION[0] == "2":
    from tensorflow.keras.models import model_from_json
else:
    from keras.models import model_from_json

import numpy as np
import time
from deepcinac.utils.utils import horizontal_flip, vertical_flip, v_h_flip
from deepcinac.cinac_movie_patch import MoviePatchGeneratorMaskedVersions, MoviePatchGeneratorForCellType, \
    MoviePatchData
# from deepcinac.utils.cells_map_utils import CellsCoord, create_cells_coord_from_suite_2p
# import PIL
import os
# from abc import ABC, abstractmethod
# from ScanImageTiffReader import ScanImageTiffReader
# from PIL import ImageSequence
import scipy.io as sio
# from shapely.geometry import MultiPoint, LineString
from datetime import datetime
import scipy.signal
import deepcinac.cinac_model
from deepcinac.utils.cinac_file_utils import CinacFileReader, read_cell_type_categories_yaml_file
from deepcinac.utils.display import plot_hist_distribution
from deepcinac.cinac_structures import create_cinac_recording_from_cinac_file_segment
import sklearn.metrics


def fusion_cell_type_predictions(cell_type_pred_1, cell_type_pred_2,
                                 cell_type_config_file=None, filename_to_save=None):
    """
    Take two cell type predictions (should be for the same movie, so the same number of cells in both),
    and fusion it so that cells that have a cell type probabilities in cell_type_pred_2 (sum of probablities equals to 1)
    will be transfered to
    cell_type_pred_1, so that the cell type for those cells in cell_type_pred_1 might change.
    If a yaml cell_type_config_file is provided, then verbose mode will be on and the changes will be print.
    It is possible to give filename_to_save and the new prediction will be saved in npy format
    Args:
        cell_type_pred_1: 2d array (n_cells, n_classes), for each cell give the probability for the cell to be one of
        the cell type (n_classes represent the number of cell type possible.
        cell_type_pred_2: 2d array (n_cells, n_classes), for each cell give the probability for the cell to be one of
        the cell type (n_classes represent the number of cell type possible.
        cell_type_config_file: (str) yaml cell type that associate a cell type name to a code (one of the column
        in cell_type_pred)
        filename_to_save: (str) should have npy extension or it will be added, should include the full path.

    Returns: a 2d array (n_cells, n_classes) representing the new cell type predictions will be returned

    """
    # first we check the that both cell pred have the same dimensions
    if cell_type_pred_1.shape[0] != cell_type_pred_2.shape[0]:
        print(f"Both cell type pred 1 & 2 should have the same number of cells instead we have "
              f"{cell_type_pred_1.shape[0]} vs {cell_type_pred_2.shape[0]}")
        return

    if cell_type_pred_1.shape[1] != cell_type_pred_2.shape[1]:
        print(f"Both cell type pred 1 & 2 should have the same number of cells type instead we have "
              f"{cell_type_pred_1.shape[1]} vs {cell_type_pred_2.shape[1]}")
        return

    # not to modify the original one
    cell_type_pred_1 = np.copy(cell_type_pred_1)

    if cell_type_config_file is not None:
        cell_type_from_code_dict, cell_type_to_code_dict, multi_class_arg = \
            read_cell_type_categories_yaml_file(yaml_file=cell_type_config_file)
        # we display info about the code:
        print(f"fusion_cell_type_predictions() on {len(cell_type_pred_1)} cells")
        for code in np.arange(cell_type_pred_1.shape[1]):
            print(f"Cell type for code {code}: {cell_type_from_code_dict[code]}")
        print(" ")
    else:
        cell_type_from_code_dict = None

    if cell_type_from_code_dict is not None:
        cell_count_by_type = np.zeros(cell_type_pred_1.shape[1], dtype="int16")
        for cell in np.arange(len(cell_type_pred_1)):
            cell_type_code = np.argmax(cell_type_pred_1[cell])
            cell_count_by_type[cell_type_code] += 1
        print(" ")
        print(f"Number of cells for each cell type before fusion:")
        for code in np.arange(len(cell_type_from_code_dict)):
            print(f"{cell_count_by_type[code]} {cell_type_from_code_dict[code]}")
        print(" ")

    n_cells_changed = 0
    n_cells_different = 0

    for cell in np.arange(len(cell_type_pred_1)):
        if np.sum(cell_type_pred_2[cell]) == 0:
            # means no predictions are given for that cell
            continue

        n_cells_changed += 1

        if cell_type_from_code_dict is not None:
            # we print some information
            print(f"Cell type pred of cell {cell} were: {np.round(cell_type_pred_1[cell], 2)} ->"
                  f"{cell_type_from_code_dict[np.argmax(cell_type_pred_1[cell])]}")

        if np.argmax(cell_type_pred_1[cell]) != np.argmax(cell_type_pred_2[cell]):
            n_cells_different += 1

        cell_type_pred_1[cell] = cell_type_pred_2[cell]
        if cell_type_from_code_dict is not None:
            print(f"New ones are: {np.round(cell_type_pred_1[cell], 2)} ->"
                  f"{cell_type_from_code_dict[np.argmax(cell_type_pred_1[cell])]}")
            print(" ")

    if cell_type_from_code_dict is not None:
        print(f"{n_cells_changed} cells had their predictions changed.")
        if n_cells_different == 0:
            print("All cells still have the same cell type.")
        elif n_cells_different == 1:
            print("1 cell has now a new cell type.")
        else:
            print(f"{n_cells_different} cells have now a news cell type.")

    if cell_type_from_code_dict is not None:
        cell_count_by_type = np.zeros(cell_type_pred_1.shape[1], dtype="int16")
        for cell in np.arange(len(cell_type_pred_1)):
            cell_type_code = np.argmax(cell_type_pred_1[cell])
            cell_count_by_type[cell_type_code] += 1
        print(" ")
        print(f"Number of cells for each cell type after fusion:")
        for code in np.arange(len(cell_type_from_code_dict)):
            print(f"{cell_count_by_type[code]} {cell_type_from_code_dict[code]}")
        print(" ")

    if filename_to_save is not None:
        if not filename_to_save.endswith(".npy"):
            filename_to_save = filename_to_save + ".npy"
        np.save(filename_to_save, cell_type_pred_1)

    return cell_type_pred_1


def select_activity_classifier_on_cell_type_outputs(cell_type_config_file, cell_type_predictions,
                                                    cell_type_to_classifier, default_classifier,
                                                    unknown_cell_type_action="ignore",
                                                    cell_type_threshold=0.5, verbose=1):
    """
    Allows to create a dictionary that will be passed to CinacPredictor.add_recording() as the model_files_dict
    argument. Each cell will be given a specific activity classifier regarding its cell type.
    Args:
        cell_type_config_file: yaml file containing the description of the cell classifier, allows
        to read the cell type predictions
        cell_type_predictions: np array of 2 dimensions, (n_cells, n_cell_type), value between 0 and 1 predicting
        if the cell is this cell type
        cell_type_threshold: use only if binary classifier, if multi_class not used; the cell_type with the maximum
        prediction is picked.
        cell_type_to_classifier: (dict), key is the name of a cell type (ex: 'interneuron') it should be in the
        yaml file, value is tuple of 3 strings (model file, weights file, classifier id) that attribute a
        specific activity classifier to a cell type
        default_classifier: (tuple of 3 strings) (model file, weights file, classifier id) that attribute a default
         activity classifier in a cell has no predominant cell type
        unknown_cell_type_action: (str) Decide what to do when cell type is not in cell_type_to_classifier. Either
        'ignore' meaning the cell with that cell type won't be predicted, "default" we use the default classifier
        on it or "exception" then an exception would be raise and the excecution will be aborded
        verbose: (int) if > 0, then the number of cells by cell type will be printed

    Returns: return a dict compatible with CinacPredictor.add_recording() model_files_dict argument.
    key is a tuple of 3 strings (model file, weights file, classifier id), and value is the cells whose activity
    should be classify using this model.

    """

    # TODO: use cell_type_threshold

    model_files_dict = dict()

    cell_type_from_code_dict, cell_type_to_code_dict, multi_class_arg = \
        read_cell_type_categories_yaml_file(yaml_file=cell_type_config_file)

    # using code instead of string
    cell_type_to_classifier_encoded = dict()
    for cell_type, model_tuple in cell_type_to_classifier.items():
        if cell_type.strip().lower() not in cell_type_to_code_dict:
            raise Exception(f"Cell type {cell_type} not recognized as a valid cell type, "
                            f"using {os.path.basename(cell_type_config_file)}")
        cell_type_code = cell_type_to_code_dict[cell_type.strip().lower()]
        cell_type_to_classifier_encoded[cell_type_code] = model_tuple

    cell_count_by_type = np.zeros(len(cell_type_from_code_dict), dtype="int16")
    for cell in np.arange(len(cell_type_predictions)):
        cell_type_code = np.argmax(cell_type_predictions[cell])
        cell_count_by_type[cell_type_code] += 1
        if len(np.where(cell_type_predictions[cell] == cell_type_predictions[cell, cell_type_code])[0]) > 1:
            # mean the cell as two cell types with the same prediction
            if default_classifier not in model_files_dict:
                model_files_dict[default_classifier] = []
            model_files_dict[default_classifier].append(cell)
        else:
            if cell_type_code not in cell_type_to_classifier_encoded:
                if unknown_cell_type_action == "ignore":
                    continue
                elif unknown_cell_type_action == "default":
                    if default_classifier not in model_files_dict:
                        model_files_dict[default_classifier] = []
                    model_files_dict[default_classifier].append(cell)
                else:
                    raise Exception(f"Cell type code {cell_type_code} not recognized as a valid cell type, "
                                f"using {os.path.basename(cell_type_config_file)}")
            else:
                model_tuple = cell_type_to_classifier_encoded[cell_type_code]
                if model_tuple not in model_files_dict:
                    model_files_dict[model_tuple] = []
                model_files_dict[model_tuple].append(cell)

    # Converting cells list as np.array
    for model_tuple, cells in model_files_dict.items():
        model_files_dict[model_tuple] = np.array(cells)

    if verbose > 0:
        # print the number of cells for each cell type
        print(" ")
        print(f"Number of cells for each cell type:")
        for code in np.arange(len(cell_type_from_code_dict)):
            print(f"{cell_count_by_type[code]} {cell_type_from_code_dict[code]}")
        print(" ")

    return model_files_dict


class CinacPredictor:

    def __init__(self, verbose=0):
        self.recordings = []
        self.verbose = verbose

    def add_recording(self, cinac_recording, model_files_dict=None, removed_cells_mapping=None):
        """
        Add a recording as a set of a calcium imaging movie, ROIs, and a list of cell to predict
        Args:
            cinac_recording: an instance of CinacRecording
            model_files_dict: dictionary. Key is a tuple of three string representing the file_name of the json file,
              the filename of the weights file and network_identifier and an identifier for the model
             and weights used (will be added to the name of files containing the predictions
             in addition of recording identifier.)
             the value of the dict being a list or array of integers that represents
             the indices of the cell to be predicted by the given model and set of weights.
             If None, means all cells activity will be predicted. Cells not including, will have their prediction
             set to 0 for all the frames
            removed_cells_mapping: integers array of length the original numbers of cells
            (such as defined in CinacRecording)
            and as value either of positive int representing the new index of the cell or -1 if the cell has been
            removed

        Returns:

        """
        self.recordings.append((cinac_recording, model_files_dict, removed_cells_mapping))

    @staticmethod
    def get_new_cell_indices_if_cells_removed(cell_indices_array, removed_cells_mapping):
        """
        Take an array of int, and return another one with new index in case some cells would have been removed
        and some cells ROIs are not matching anymore
        Args:
            cell_indices_array: np.array of integers, containing the cells indices to remap
            removed_cells_mapping: integers array of length the original numbers of cells
            (such as defined in CinacRecording)
            and as value either of positive int representing the new index of the cell or -1 if the cell has been
            removed

        Returns: new_cell_indices_array an np.array that contains integers representing the new indices of cells that
        have not been removed
        original_cell_indices_mapping, np.array, for each new cell index, contains the corresponding original index

        """

        if removed_cells_mapping is None:
            return np.copy(cell_indices_array), np.copy(cell_indices_array)

        new_cell_indices_array = removed_cells_mapping[cell_indices_array]
        # removing cell indices of cell that has been removed
        copy_new_cell_indices_array = np.copy(new_cell_indices_array)
        new_cell_indices_array = new_cell_indices_array[new_cell_indices_array >= 0]
        original_cell_indices_mapping = np.copy(cell_indices_array[copy_new_cell_indices_array >= 0])

        return new_cell_indices_array, original_cell_indices_mapping

    def predict(self, results_path, overlap_value=0.5, output_file_formats="npy", cell_type_classifier_mode=False,
                n_segments_to_use_for_prediction=2, cell_type_pred_fct=np.mean, create_dir_for_results=True,
                cell_buffer=None, all_pixels=True,
                **kwargs):
        """
        Will predict the neural activity state of the cell for each frame of the calcium imaging movies.
        Recordings have to be added previously though the add_recording method.
        Args:
            results_path: if None, predictions are not saved in a file
            overlap_value:
            cell_type_classifier_mode: means we want to classify each cell to know its type. So far return a prediction
            value between 0 and 1, 1 meaning 100% a pyramidal cell, 0 100% sure not (should be interneuron then)
            output_file_formats: a string or list of string, representing the format in which saving the 2d array
            representing the prediction for each movie. The choices are: "mat" or "npy" or "npz". npy and npz will both
            comes to a npz format. Two keys will be in the .mat or .npz, one is "predictions" containing a 1d or 2 array
            with predictions inside (1d array for cell activity predictions, 2d array for cell type activity predicctions
            (n_cells, n_cells_type), and the second key is "cells" which is a 1d array containing the indices of
            the cells that has been predicted (other cells will have a 0 value prediction in the predictions array).
            n_segments_to_use_for_prediction: used when cell_type_classifier_mode is at True. Indicate how many segment
            of window_len should be used to predict the cell_type. For example if the value is 2 and window_len is 200,
            then we take the mean of the prediction over 2 segments of 200 frames. Those segments are selected based
            on their activity (the more a cell is active (amplitudes) the best chance the segment is to be selected)
            cell_type_pred_fct: fct to use to average the predictons if it is done in more than one segment
            create_dir_for_results: if True, create a dir with timestamps to store the results. The directory is created
            in results_path. Else predictions are stored directly in results_path
            cell_buffer: None or int, distance in order to transform the cell contour giving an approximate
            representation of  all points within a given distance of the this geometric object
            all_pixels: bool, use for cell type prediction, add input with all pixels
            **kwargs:


        Returns: a dict with key being the id of the cinac recording and value:
        numpy array of n_cells containing the predictions (1d for cell activity, and 2d for cell type
        (n_cells, n_cell_types),.

        """

        use_data_augmentation = False
        if "use_data_augmentation" in kwargs:
            use_data_augmentation = kwargs["use_data_augmentation"]

        # output_file_formats = "npy"
        if "output_file_formats" in kwargs:
            output_file_formats = kwargs["output_file_formats"]

        # dictionary that will contains the predictions results. Keys are the CinacRecording identifiers
        predictions_dict = dict()
        for recording in self.recordings:
            cinac_recording, model_files_dict, removed_cells_mapping = recording
            n_cells = cinac_recording.get_n_cells()
            n_frames = cinac_recording.get_n_frames()

            network_identifiers = []
            for file_names, cells_to_predict in model_files_dict.items():
                json_model_file_name, weights_file_name, network_identifier = file_names
                network_identifiers.append(network_identifier)
                # ms.tiffs_for_transient_classifier_path = tiffs_for_transient_classifier_path

                if cells_to_predict is None:
                    cells_to_load = np.arange(n_cells)
                else:
                    cells_to_load = np.array(cells_to_predict)
                    # print(f"cells_to_load {cells_to_load}")
                    # if cell_type_classifier_mode:
                    #     n_cells = len(cells_to_load)

                print(f"{len(cells_to_load)} cells to predict with {network_identifier}")

                original_cells_to_load = np.copy(cells_to_load)
                cells_to_load, \
                original_cell_indices_mapping = self.get_new_cell_indices_if_cells_removed(cells_to_load,
                                                                                           removed_cells_mapping)

                total_n_cells = len(cells_to_load)
                # print(f'total_n_cells {total_n_cells}')

                if total_n_cells == 0:
                    print(f"No cells loaded for {cinac_recording.identifier}")
                    continue

                # print(f"transients_prediction_from_movie n_frames {n_frames}")
                # we keep the original number of cells, so if a different segmentation was used another prediction,
                # we can still compare it using the indices we know
                if cell_type_classifier_mode:
                    # will be initiated later when the number of classes will be known
                    predictions_by_cell = None  # np.zeros(n_cells)
                else:
                    predictions_by_cell = np.zeros((n_cells, n_frames))

                # loading model
                # Model reconstruction from JSON file
                with open(json_model_file_name, 'r') as f:
                    model = model_from_json(f.read(), custom_objects={"deepcinac.cinac_model": deepcinac.cinac_model})

                # Load weights into the new model
                model.load_weights(weights_file_name)

                start_time = time.time()
                predictions_threshold = 0.5
                # in case we want to use tqdm
                # for cell_index in tqdm(range(len(cells_to_load)),
                #                              desc=f"{cinac_recording.identifier} - {network_identifier}"):
                #     cell = cells_to_load[cell_index]
                for cell_index in range(len(cells_to_load)):
                    cell = cells_to_load[cell_index]
                    original_cell = original_cell_indices_mapping[cell_index]
                    if cell_type_classifier_mode:
                        predictions = predict_cell_type_from_model(cinac_recording=cinac_recording,
                                                                   cell=cell, model=model, overlap_value=overlap_value,
                                                                   use_data_augmentation=use_data_augmentation,
                                                                   n_frames=n_frames,
                                                                   all_pixels=all_pixels,
                                                                   n_segments_to_use_for_prediction=
                                                                   n_segments_to_use_for_prediction,
                                                                   verbose=self.verbose)
                        # print(f"predictions in predict() {len(predictions)} {len(predictions[0])} {predictions}")
                        if len(predictions) > 1:
                            if len(predictions[0]) == 1:
                                if predictions_by_cell is None:
                                    predictions_by_cell = np.zeros(n_cells)
                                predictions = cell_type_pred_fct(predictions)
                            else:
                                if predictions_by_cell is None:
                                    predictions_by_cell = np.zeros((n_cells, len(predictions[0])))
                                predictions = cell_type_pred_fct(predictions, axis=0)
                        else:
                            if predictions_by_cell is None:
                                predictions_by_cell = np.zeros(n_cells)
                            predictions = predictions[0]
                        predictions_by_cell[original_cell] = predictions
                    else:
                        predictions = predict_transient_from_model(cinac_recording=cinac_recording,
                                                                   cell=cell, model=model, overlap_value=overlap_value,
                                                                   use_data_augmentation=use_data_augmentation,
                                                                   n_frames=n_frames, buffer=cell_buffer, pixels_around=0)
                        # print(f"### cinac_predictor predict(): np.sum(predictions) {np.sum(predictions)}")
                        if len(predictions.shape) == 1:
                            predictions_by_cell[original_cell] = predictions
                        elif (len(predictions.shape) == 2) and (predictions.shape[1] == 1):
                            predictions_by_cell[original_cell] = predictions[:, 0]
                        elif (len(predictions.shape) == 2) and (predictions.shape[1] == 3):
                            # real transient, fake ones, other (neuropil, decay etc...)
                            # keeping predictions about real transient when superior
                            # to other prediction on the same frame
                            max_pred_by_frame = np.max(predictions, axis=1)
                            real_transient_frames = (predictions[:, 0] == max_pred_by_frame)
                            predictions_by_cell[original_cell, real_transient_frames] = 1
                        elif predictions.shape[1] == 2:
                            # real transient, fake ones
                            # keeping predictions about real transient superior to the threshold
                            # and superior to other prediction on the same frame
                            max_pred_by_frame = np.max(predictions, axis=1)
                            real_transient_frames = np.logical_and((predictions[:, 0] >= predictions_threshold),
                                                                   (predictions[:, 0] == max_pred_by_frame))
                            predictions_by_cell[original_cell, real_transient_frames] = 1

            stop_time = time.time()
            if self.verbose > 0:
                print(f"Time to predict {total_n_cells} cells: "
                      f"{np.round(stop_time - start_time, 3)} s")
            if isinstance(output_file_formats, str):
                output_file_formats = [output_file_formats]

            # TODO: use all network_identifier if more than one
            if len(network_identifiers) == 0:
                network_identifier = ""
            else:
                network_identifier = "_" + "_".join(network_identifiers)
            if results_path is not None:
                if create_dir_for_results:
                    time_str = datetime.now().strftime("%Y_%m_%d.%H-%M-%S")
                    new_dir_results = os.path.join(results_path, f"{time_str}/")
                    os.mkdir(new_dir_results)
                else:
                    new_dir_results = results_path

                predictions_str = "predictions"
                if cell_type_classifier_mode:
                    predictions_str = "cell_type_predictions"
                file_name = f"{cinac_recording.identifier}_{predictions_str}{network_identifier}"
                for output_file_format in output_file_formats:
                    if "mat" in output_file_format:
                        sio.savemat(os.path.join(new_dir_results, file_name + ".mat"),
                                    {'predictions': predictions_by_cell, "cells": original_cells_to_load})
                    elif ("npy" in output_file_format) or ("npz" in output_file_format):
                        np.savez(os.path.join(new_dir_results, file_name + ".npz"),
                                 predictions=predictions_by_cell, cells=original_cells_to_load)
            predictions_dict[cinac_recording.identifier] = predictions_by_cell

        return predictions_dict


def predict_cell_type_from_model(cinac_recording, cell, model,
                                 n_frames, n_segments_to_use_for_prediction=2,
                                 overlap_value=0, pixels_around=0, all_pixels=False,
                                 use_data_augmentation=False, buffer=None, verbose=0):
    # if n_frames is None, then the movie need to have been loaded
    # start_time = time.time()
    # multi_inputs = (model.layers[0].output_shape == model.layers[1].output_shape)
    # print(f"model.layers[0].output_shape {model.layers[0].output_shape}")
    window_len = model.layers[0].output_shape[0][1]
    max_height = model.layers[0].output_shape[0][2]
    max_width = model.layers[0].output_shape[0][3]

    # Determining how many classes were used
    if len(model.layers[-1].output_shape) == 2:
        using_multi_class = 1
    else:
        using_multi_class = model.layers[-1].output_shape[2]
        # print(f"predict_transient_from_model using_multi_class {using_multi_class}")

    if use_data_augmentation:
        augmentation_functions = [horizontal_flip, vertical_flip, v_h_flip]
    else:
        augmentation_functions = None

    movie_patches, data_frame_indices = load_data_for_cell_type_prediction(cinac_recording=cinac_recording,
                                                                           cell=cell,
                                                                           sliding_window_len=window_len,
                                                                           overlap_value=overlap_value,
                                                                           augmentation_functions=augmentation_functions,
                                                                           n_frames=n_frames,
                                                                           n_windows_len_to_keep_by_cell=
                                                                           n_segments_to_use_for_prediction)

    movie_patch_generator = \
        MoviePatchGeneratorForCellType(window_len=window_len, max_width=max_width, max_height=max_height,
                                       pixels_around=pixels_around, buffer=buffer, with_all_pixels=all_pixels,
                                       using_multi_class=using_multi_class, cell_type_classifier_mode=True)

    # print(f"Cell {cell} from {cinac_recording.identifier}")
    data_dict = movie_patch_generator.generate_movies_from_metadata(movie_data_list=movie_patches,
                                                                    with_labels=False)

    # stop_time = time.time()
    # print(f"Time to get the data: "
    #       f"{np.round(stop_time - start_time, 3)} s")
    start_time = time.time()

    predictions = model.predict(data_dict)

    stop_time = time.time()
    if verbose > 0:
        print(f"Time to get predictions for cell {cell}: "
              f"{np.round(stop_time - start_time, 3)} s")

    return predictions


def load_data_for_prediction(cinac_recording, cell, sliding_window_len, overlap_value,
                             augmentation_functions, n_frames):
    """
    Create data that will be used to predict neural activity. The data will be representing by instances of
    MoviePatchData that will contains information concerning this movie segment for a given
    cell to give to the neuronal network.
    Args:
        cinac_recording: instance of CinacRecording, contains the movie frames and the ROIs
        cell: integer, the cell index
        sliding_window_len: integer, length in frames of the window used by the neuronal network. Predictions will be
        made for each frame of this segment
        overlap_value: float value between 0 and 1, representing by how much 2 movie segments will overlap.
        0.5 is equivalent to a 50% overlap. It allows the network to avoid edge effect in order to get a full temporal
        vision of all transient. A good default value is 0.5
        augmentation_functions: list of function that takes an image (np array) as input and return a copy of the image
        transformed.
        n_frames: number of frames in the movie

    Returns: a list of MoviePatchData instance and an integer representing the index of the first frame of the patch

    """
    # n_frames is None, the movie need to have been loaded
    # we suppose that the movie is already loaded and normalized
    movie_patches = []
    data_frame_indices = []
    frames_step = int(np.ceil(sliding_window_len * (1 - overlap_value)))
    # number of indices to remove so index + sliding_window_len won't be superior to number of frames
    n_step_to_remove = 0 if (overlap_value == 0) else int(1 / (1 - overlap_value))
    frame_indices_for_movies = np.arange(0, n_frames, frames_step)
    if n_step_to_remove > 0:
        frame_indices_for_movies = frame_indices_for_movies[:-n_step_to_remove + 1]
    # in case the n_frames wouldn't be divisible by frames_step
    if frame_indices_for_movies[-1] + frames_step > n_frames:
        frame_indices_for_movies[-1] = n_frames - sliding_window_len

    for i, index_movie in enumerate(frame_indices_for_movies):
        break_it = False
        first_frame = index_movie
        if (index_movie + sliding_window_len) == n_frames:
            break_it = True
        elif (index_movie + sliding_window_len) > n_frames:
            # in case the number of frames is not divisible by sliding_window_len
            first_frame = n_frames - sliding_window_len
            break_it = True
        movie_data = MoviePatchData(cinac_recording=cinac_recording, cell=cell, index_movie=first_frame,
                                    window_len=sliding_window_len,
                                    max_n_transformations=3,
                                    cell_type_classifier_mode=False,
                                    with_info=False, encoded_frames=None,
                                    decoding_frame_dict=None)

        movie_patches.append(movie_data)
        data_frame_indices.append(first_frame)
        if augmentation_functions is not None:
            for augmentation_fct in augmentation_functions:
                new_movie = movie_data.copy()
                new_movie.data_augmentation_fct = augmentation_fct
                movie_patches.append(new_movie)
                data_frame_indices.append(first_frame)

        if break_it:
            break

    return movie_patches, data_frame_indices


def load_data_for_cell_type_prediction(cinac_recording, cell, sliding_window_len, overlap_value,
                                       augmentation_functions, n_frames, n_windows_len_to_keep_by_cell=2):
    """
    Create data that will be used to predict cell type. The data will be representing by instances of
    MoviePatchData that will contains information concerning this movie segment for a given
    cell to give to the neuronal network.
    Args:
        cinac_recording: instance of CinacRecording, contains the movie frames and the ROIs
        cell: integer, the cell index
        sliding_window_len: integer, length in frames of the window used by the neuronal network. Predictions will be
        made for each frame of this segment
        overlap_value: float value between 0 and 1, representing by how much 2 movie segments will overlap.
        0.5 is equivalent to a 50% overlap. It allows the network to avoid edge effect in order to get a full temporal
        vision of all transient. A good default value is 0.5
        augmentation_functions: list of function that takes an image (np array) as input and return a copy of the image
        transformed.
        n_frames: number of frames in the movie
        n_windows_len_to_keep_by_cell: how many segment of window_len frames should be used to predict cell_type,
        so the length of the returned list will be the same as n_windows_len_to_keep_by_cell
    Returns: a list of MoviePatchData instance and an integer representing the index of the first frame of the patch

    """
    # n_frames is None, the movie need to have been loaded
    # we suppose that the movie is already loaded and normalized
    movie_patches = []
    data_frame_indices = []
    frames_step = int(np.ceil(sliding_window_len * (1 - overlap_value)))
    # number of indices to remove so index + sliding_window_len won't be superior to number of frames
    n_step_to_remove = 0 if (overlap_value == 0) else int(1 / (1 - overlap_value))
    frame_indices_for_movies = np.arange(0, n_frames, frames_step)
    if n_step_to_remove > 0:
        frame_indices_for_movies = frame_indices_for_movies[:-n_step_to_remove + 1]
    # in case the n_frames wouldn't be divisible by frames_step
    if frame_indices_for_movies[-1] + frames_step > n_frames:
        frame_indices_for_movies[-1] = n_frames - sliding_window_len

    # list of  int representing the first_frame possible candidate
    # for moviepatch to add for predicting. We want to keep for each cell only the most active
    # one
    first_frames_to_filter = []
    n_windows_len_to_keep_by_cell = max(0, n_windows_len_to_keep_by_cell)

    for i, index_movie in enumerate(frame_indices_for_movies):
        break_it = False
        first_frame = index_movie
        if (index_movie + sliding_window_len) == n_frames:
            break_it = True
        elif (index_movie + sliding_window_len) > n_frames:
            # in case the number of frames is not divisible by sliding_window_len
            first_frame = n_frames - sliding_window_len
            break_it = True
        first_frames_to_filter.append(first_frame)

        if break_it:
            break

    # we want to select at least a first_frame to create a moviePatch
    # the idea is to sort the first_frame and the segment that goes with (adding self.window_len)
    # according to the activity of the cell, we're looking for the segment with the highest amplitude peaks
    first_frames_selected = []
    # will contain a score representing the avg amplitude of the highest peaks
    avg_peaks_amplitudes = []
    smooth_traces = cinac_recording.get_smooth_traces(normalized=True)
    if len(smooth_traces.shape) == 1:
        pass
    else:
        smooth_traces = smooth_traces[cell]

    peaks, properties = scipy.signal.find_peaks(x=smooth_traces, distance=2)
    peak_nums = np.zeros(n_frames, dtype="int8")
    peak_nums[peaks] = 1
    for first_frame in first_frames_to_filter:
        peaks = np.where(peak_nums[first_frame:first_frame + sliding_window_len])[0] + first_frame
        if len(peaks) == 0:
            # no peaks found
            avg_peaks_amplitudes.append(0)
        amplitudes = smooth_traces[peaks]
        # print(f"len(amplitudes) {len(amplitudes)}")
        # print(f"amplitudes {amplitudes}")
        amplitude_threshold = np.percentile(amplitudes, 90)
        highest_amplitudes = amplitudes[amplitudes >= amplitude_threshold]
        # print(f"len(highest_amplitudes) {len(highest_amplitudes)}, "
        #       f"mean {np.round(np.mean(highest_amplitudes), 2)}")
        avg_peaks_amplitudes.append(np.mean(highest_amplitudes))
    first_frames_order = np.argsort(avg_peaks_amplitudes)
    # from highest to the smallest
    first_frames_order = first_frames_order[::-1]
    # print(f"avg_peaks_amplitudes ordered {np.array(avg_peaks_amplitudes)[first_frames_order]}")
    n_first_frames_to_peak = min(len(first_frames_order), n_windows_len_to_keep_by_cell)
    for index_first_frame in first_frames_order[:n_first_frames_to_peak]:
        first_frames_selected.append(first_frames_to_filter[index_first_frame])

    for first_frame in first_frames_selected:
        movie_data = MoviePatchData(cinac_recording=cinac_recording, cell=cell, index_movie=first_frame,
                                    window_len=sliding_window_len,
                                    max_n_transformations=3,
                                    with_info=False, encoded_frames=None,
                                    cell_type_classifier_mode=True,
                                    decoding_frame_dict=None)

        movie_patches.append(movie_data)
        data_frame_indices.append(first_frame)
        if augmentation_functions is not None:
            for augmentation_fct in augmentation_functions:
                new_movie = movie_data.copy()
                new_movie.data_augmentation_fct = augmentation_fct
                movie_patches.append(new_movie)
                data_frame_indices.append(first_frame)

    return movie_patches, data_frame_indices


def predict_transient_from_model(cinac_recording, cell, model,
                                 n_frames, overlap_value=0.8, pixels_around=0,
                                 use_data_augmentation=False, buffer=None):
    # if n_frames is None, then the movie need to have been loaded
    # start_time = time.time()
    # multi_inputs = (model.layers[0].output_shape == model.layers[1].output_shape)
    if len(model.layers[0].output_shape) == 1:
        window_len = model.layers[0].output_shape[0][1]
        max_height = model.layers[0].output_shape[0][2]
        max_width = model.layers[0].output_shape[0][3]
    else:
        window_len = model.layers[0].output_shape[1]
        max_height = model.layers[0].output_shape[2]
        max_width = model.layers[0].output_shape[3]

    # Determining how many classes were used
    if len(model.layers[-1].output_shape) == 2:
        using_multi_class = 1
    else:
        using_multi_class = model.layers[-1].output_shape[2]
        # print(f"predict_transient_from_model using_multi_class {using_multi_class}")

    if use_data_augmentation:
        augmentation_functions = [horizontal_flip, vertical_flip, v_h_flip]
    else:
        augmentation_functions = None
    # start_time_bis = time.time()
    movie_patches, data_frame_indices = load_data_for_prediction(cinac_recording=cinac_recording,
                                                                 cell=cell,
                                                                 sliding_window_len=window_len,
                                                                 overlap_value=overlap_value,
                                                                 augmentation_functions=augmentation_functions,
                                                                 n_frames=n_frames)
    # stop_time_bis = time.time()
    # print(f"Time to get load_data_for_prediction: "
    #       f"{np.round(stop_time_bis - start_time_bis, 3)} s")

    movie_patch_generator = \
        MoviePatchGeneratorMaskedVersions(window_len=window_len, max_width=max_width, max_height=max_height,
                                          pixels_around=pixels_around, buffer=buffer, with_neuropil_mask=True,
                                          using_multi_class=using_multi_class, cell_type_classifier_mode=False)

    data_dict = movie_patch_generator.generate_movies_from_metadata(movie_data_list=movie_patches,
                                                                    with_labels=False)

    # stop_time = time.time()
    # print(f"Time to get the data: "
    #       f"{np.round(stop_time - start_time, 3)} s")
    start_time = time.time()

    predictions = model.predict(data_dict)

    stop_time = time.time()
    print(f"Time to get predictions for cell {cell}: "
          f"{np.round(stop_time - start_time, 3)} s")

    # now we want to take the prediction with the maximal value for a given frame
    # the rational being that often if a transient arrive right at the end or the beginning
    # it won't be recognized as True
    if (overlap_value > 0) or (augmentation_functions is not None):
        frames_predictions = dict()
        # print(f"predictions.shape {predictions.shape}, data_frame_indices.shape {data_frame_indices.shape}")
        for i, data_frame_index in enumerate(data_frame_indices):
            frames_index = np.arange(data_frame_index, data_frame_index + window_len)
            predictions_for_frames = predictions[i]
            for j, frame_index in enumerate(frames_index):
                if frame_index not in frames_predictions:
                    frames_predictions[frame_index] = dict()
                if len(predictions_for_frames.shape) == 1:
                    if 0 not in frames_predictions[frame_index]:
                        frames_predictions[frame_index][0] = []
                    frames_predictions[frame_index][0].append(predictions_for_frames[j])
                else:
                    # then it's muti_class labels
                    for index in np.arange(len(predictions_for_frames[j])):
                        if index not in frames_predictions[frame_index]:
                            frames_predictions[frame_index][index] = []
                        frames_predictions[frame_index][index].append(predictions_for_frames[j, index])

        predictions = np.zeros((n_frames, using_multi_class))
        for frame_index, class_dict in frames_predictions.items():
            for class_index, prediction_values in class_dict.items():
                # max value
                predictions[frame_index, class_index] = np.max(prediction_values)
    else:
        # to flatten all but last dimensions
        predictions = np.ndarray.flatten(predictions)

        # now we remove the extra prediction in case the number of frames was not divisible by the window length
        if (n_frames % window_len) != 0:
            print("(n_frames % window_len) != 0")
            real_predictions = np.zeros((n_frames, using_multi_class))
            modulo = n_frames % window_len
            real_predictions[:len(predictions) - window_len] = predictions[
                                                               :len(predictions) - window_len]
            real_predictions[len(predictions) - window_len:] = predictions[-modulo:]
            predictions = real_predictions

    if len(predictions) != n_frames:
        raise Exception(f"predictions len {len(predictions)} is different from the number of frames {n_frames}")

    return predictions


def activity_predictions_from_cinac_files(cinac_dir_name, results_path,
                                          json_file_name, weights_file_name,
                                          classifier_id=None,
                                          output_file_formats="npy"):
    """
        Evaluate activity prediction on cinac file with ground truth.
        Load all cinac files from a folder and compare the predictions from the classifier to the activity in the cinac
        file and then display the metrics and can save the plot of the distribution of the probability by cell type.
        Args:
            cinac_dir_name: Directory in which to find the .cinac file to use
            results_path: Directory in which to save the plots, can be None if save_activity_distribution is False
            json_file_name: .json file containing the model to use to classify the cell types
            weights_file_name: .h5 file containing the weights of the network to classify the cell types
            save_activity_distribution: (bool) if True, a figure representing the distribution of the predictions score
            for activity will be save in results_path

        Returns:

    """
    cinac_path_w_file_names = []
    # frames_tp = 0
    # frames_tn = 0
    # frames_fp = 0
    # frames_fn = 0
    #
    # transients_tp = 0
    # transients_tn = 0
    # transients_fp = 0
    # transients_fn = 0

    # look for filenames in the fist directory, if we don't break, it will go through all directories
    for (dirpath, dirnames, local_filenames) in os.walk(cinac_dir_name):
        cinac_path_w_file_names = [os.path.join(dirpath, f) for f in local_filenames if f.endswith(".cinac")]
        cinac_file_names = [f for f in local_filenames if f.endswith(".cinac")]
        break

    for file_index, cinac_file_name in enumerate(cinac_path_w_file_names):
        cinac_file_reader = CinacFileReader(file_name=cinac_file_name)
        n_cells = cinac_file_reader.get_n_cells()
        if n_cells is None:
            # then we can't built the predicted raster with all cells segmented
            print(f"The number of segmented cells in {cinac_file_reader.base_name} is not available")
            continue
        n_frames = cinac_file_reader.get_n_frames()
        if n_frames is None:
            # then we can't built the predicted raster with all cells segmented
            print(f"The number of frames in {cinac_file_reader.base_name} is not available")
            continue
        predictions_for_all_cells = np.zeros((n_cells, n_frames))
        # cinac_movie = get_cinac_movie_from_cinac_file_reader(cinac_file_reader)
        segments_list = cinac_file_reader.get_all_segments()
        # identifier = os.path.basename(cinac_file_name)
        identifier = cinac_file_reader.base_name
        for segment in segments_list:
            cell = segment[0]
            first_frame = segment[1]
            last_frame = segment[2]
            print(f"Evaluating {cinac_file_reader.base_name}, cell {segment[0]}: {segment[1]} - {segment[2]}")
            cinac_recording = create_cinac_recording_from_cinac_file_segment(identifier=identifier,
                                                                             cinac_file_reader=cinac_file_reader,
                                                                             segment=segment)

            """
                Then we decide which network will be used for predicting the cells' activity.

                A dictionnary with key a tuple of 3 elements is used.

                The 3 elements are:

                (string) the model file name (.json extension)
                (string) the weights of the network file name (.h5 extension)
                (string) identifier for this configuration, will be used to name the output file
                The dictionnary will contain as value the cells to be predicted by the key configuration. 
                If the value is set to None, then all the cells will be predicted using this configuration.
                """

            model_files_dict = dict()
            # we just one to predict the cell of interest, which is the first cell
            model_files_dict[(json_file_name, weights_file_name, identifier)] = np.arange(1)

            """
                        We now create an instance of CinacPredictor and add the CinacRecording we have just created.

                        It's possible to add more than one instance of CinacRecording, they will be predicted on the same run then.

                        The argument removed_cells_mapping allows to remove cells from the segmentation. 
                        This could be useful as the network take in consideration the adjacent cells to predict the activity, 
                        thus if a cell was wrongly added to segmentation, this could lower the accuracy of the classifier.
                        """

            cinac_predictor = CinacPredictor()

            """
            Args:

                removed_cells_mapping: integers array of length the original numbers of 
                    cells (such as defined in CinacRecording)
                    and as value either of positive int representing the new index of 
                    the cell or -1 if the cell has been removed
            """

            cinac_predictor.add_recording(cinac_recording=cinac_recording,
                                          removed_cells_mapping=None,
                                          model_files_dict=model_files_dict)

            """
            Finally, we run the prediction.

            The output format could be either a matlab file(.mat) and/or numpy one (.npy).

            If matlab is chosen, the predictions will be available under the key "predictions".

            The predictions are a 2d float array (n_cells * n_frames) with value between 0 and 1, 
            representing the prediction of our classifier for each frame. 
            1 means the cell is 100% sure active at that time, 0 is 100% sure not active.

            A cell is considered active during the rising time of the calcium transient.

            We use a threshold of 0.5 to binarize the predictions array and make it a raster.
            """

            # you could decomment this line to make sure the GPU is used
            # with tf.device('/device:GPU:0'):

            # predictions are saved in the results_path and return as a dict,
            predictions_dict = cinac_predictor.predict(results_path=None, overlap_value=0.5, output_file_formats="npy",
                                                       cell_type_classifier_mode=False,
                                                       n_segments_to_use_for_prediction=2,
                                                       create_dir_for_results=False)
            predictions_for_cinac_file = predictions_dict[list(predictions_dict.keys())[0]]
            # predictions for the first cell
            # prediction_valuesis a 1d array of len n_frames in the segment
            prediction_values = predictions_for_cinac_file[0]
            predictions_for_all_cells[cell, first_frame:last_frame+1] = prediction_values
            # we build the raster, putting 1 at every frame in which the cell is predicted as active
            # predicted_raster_dur = np.zeros(len(prediction_values), dtype="int8")
            # predicted_raster_dur[prediction_values > 0.5] = 1
        file_name = f"{identifier}_predictions"
        if classifier_id is not None:
            file_name = file_name + "_" + classifier_id
        if isinstance(output_file_formats, str):
            output_file_formats = [output_file_formats]
        # TODO: see to save only the cells predicted; adding a field cells that give the indices of the cells predicted
        for output_file_format in output_file_formats:
            if "mat" in output_file_format:
                sio.savemat(os.path.join(results_path, file_name + ".mat"),
                            {'predictions': predictions_for_all_cells})
            elif "npz" in output_file_format:
                np.savez(os.path.join(results_path, file_name + ".npz"),
                         predictions=predictions_for_all_cells)
            elif "npy" in output_file_format:
                np.save(os.path.join(results_path, file_name + ".npy"), predictions_for_all_cells)


def evaluate_cell_type_predictions(cinac_dir_name, cell_type_yaml_file, results_path,
                                   json_file_name, weights_file_name, save_cell_type_distribution, all_pixels=False
                                   ):
    """
    Evaluation the cell type prediction on cinac file with ground truth.
    Load all cinac files from a folder and compare the predictions from the classifier to the cell type in the cinac
    file and then display the metrics and can save the plot of the distribution of the probability by cell type.
    Args:
        cinac_dir_name: Directory in which to find the .cinac file to use
        cell_type_yaml_file: ymal file containing the configuration to use by the classifier
        (cell type, number of classes)
        results_path: Directory in which to save the plots, can be None if save_cell_type_distribution is False
        json_file_name: .json file containing the model to use to classify the cell types
        weights_file_name: .h5 file containing the weights of the network to classify the cell types
        save_cell_type_distribution: (bool) if True, a figure representing the distribution of the predictions score
        of each cell type will be save in results_path
        all_pixels: if True, add an input with all pixels in the frame

    Returns:

    """
    cell_type_from_code_dict, cell_type_to_code_dict, multi_class_arg = \
        read_cell_type_categories_yaml_file(yaml_file=cell_type_yaml_file, using_multi_class=1)

    print(f"### cell_type_from_code_dict {cell_type_from_code_dict}")

    n_cell_categories = len(cell_type_from_code_dict)

    if n_cell_categories < 2:
        raise Exception(f"You need at least 2 cell_type categories, you provided {n_cell_categories}: "
                        f"{list(cell_type_from_code_dict.values())}")

    # if n_class_for_predictions is equal 1, it means we are building a binary classifier
    # otherwise a multi class one

    if multi_class_arg is not None:
        if multi_class_arg or (n_cell_categories > 2):
            n_class_for_predictions = n_cell_categories
        else:
            n_class_for_predictions = 1
    else:
        if n_cell_categories > 2:
            n_class_for_predictions = n_cell_categories
        else:
            n_class_for_predictions = 1

    cinac_path_w_file_names = []
    cinac_file_names = []
    tp = 0
    tn = 0
    fp = 0
    fn = 0
    # used for multiclass
    more_than_one_cell_type = 0
    none_of_cell_types = 0
    # for sklearn metrics
    # from https://towardsdatascience.com/multi-class-metrics-made-simple-part-i-precision-and-recall-9250280bddc2
    y_true = []
    y_pred = []
    # each key is the code representing the cell_type (see in cell_type_from_code_dict), each value is an int
    # representing the number of cell by type
    n_cells_by_type_dict = dict()
    cell_type_predictions_dict = dict()
    # look for filenames in the fist directory, if we don't break, it will go through all directories
    for (dirpath, dirnames, local_filenames) in os.walk(cinac_dir_name):
        cinac_path_w_file_names = [os.path.join(dirpath, f) for f in local_filenames if f.endswith(".cinac")]
        cinac_file_names = [f for f in local_filenames if f.endswith(".cinac")]
        break
    for file_index, cinac_file_name in enumerate(cinac_path_w_file_names):
        cinac_file_reader = CinacFileReader(file_name=cinac_file_name)
        # cinac_movie = get_cinac_movie_from_cinac_file_reader(cinac_file_reader)
        segments_list = cinac_file_reader.get_all_segments()
        # identifier = os.path.basename(cinac_file_name)
        identifier = cinac_file_names[file_index]
        for segment in segments_list:
            cinac_recording = create_cinac_recording_from_cinac_file_segment(identifier=identifier,
                                                                             cinac_file_reader=cinac_file_reader,
                                                                             segment=segment)

            """
                Then we decide which network will be used for predicting the cells' activity.

                A dictionnary with key a tuple of 3 elements is used.

                The 3 elements are:

                (string) the model file name (.json extension)
                (string) the weights of the network file name (.h5 extension)
                (string) identifier for this configuration, will be used to name the output file
                The dictionnary will contain as value the cells to be predicted by the key configuration. 
                If the value is set to None, then all the cells will be predicted using this configuration.
                """

            model_files_dict = dict()
            # we just one to predict the cell of interest, which is the first cell
            model_files_dict[(json_file_name, weights_file_name, identifier)] = np.arange(1)

            """
            We now create an instance of CinacPredictor and add the CinacRecording we have just created.

            It's possible to add more than one instance of CinacRecording, they will be predicted on the same run then.

            The argument removed_cells_mapping allows to remove cells from the segmentation. 
            This could be useful as the network take in consideration the adjacent cells to predict the activity, 
            thus if a cell was wrongly added to segmentation, this could lower the accuracy of the classifier.
            """

            cinac_predictor = CinacPredictor()

            """
            Args:

                removed_cells_mapping: integers array of length the original numbers of 
                    cells (such as defined in CinacRecording)
                    and as value either of positive int representing the new index of 
                    the cell or -1 if the cell has been removed
            """

            cinac_predictor.add_recording(cinac_recording=cinac_recording,
                                          removed_cells_mapping=None,
                                          model_files_dict=model_files_dict)

            """
            Finally, we run the prediction.

            The output format could be either a matlab file(.mat) and/or numpy one (.npy).

            If matlab is chosen, the predictions will be available under the key "predictions".

            The predictions are a 2d float array (n_cells * n_frames) with value between 0 and 1, representing the prediction of our classifier for each frame. 1 means the cell is 100% sure active at that time, 0 is 100% sure not active.

            A cell is considered active during the rising time of the calcium transient.

            We use a threshold of 0.5 to binarize the predictions array and make it a raster.
            """

            # you could decomment this line to make sure the GPU is used
            # with tf.device('/device:GPU:0'):

            # predictions are saved in the results_path and return as a dict,
            predictions_dict = cinac_predictor.predict(results_path=None, output_file_formats="npy",
                                                       overlap_value=0, cell_type_classifier_mode=True,
                                                       n_segments_to_use_for_prediction=4,
                                                       cell_type_pred_fct=np.mean,
                                                       create_dir_for_results=False,
                                                       all_pixels=all_pixels)
            # mean of the different predictions is already in the dict
            prediction_value = predictions_dict[list(predictions_dict.keys())[0]][0]

            right_prediction = False

            if cinac_recording.cell_type.strip().lower() not in cell_type_to_code_dict:
                print(f"cinac_recording.cell_type ({cinac_recording.cell_type}) not in cell_type_to_code_dict")
                continue
            cell_type_code = cell_type_to_code_dict[cinac_recording.cell_type.strip().lower()]
            if cell_type_code not in n_cells_by_type_dict:
                n_cells_by_type_dict[cell_type_code] = 0
                cell_type_predictions_dict[cell_type_code] = []
            n_cells_by_type_dict[cell_type_code] += 1
            if n_class_for_predictions == 1:
                cell_type_predictions_dict[cell_type_code].append(prediction_value)
                if cell_type_code == 0:
                    if prediction_value >= 0.5:
                        fp += 1
                    else:
                        tn += 1
                        right_prediction = True
                else:
                    if prediction_value >= 0.5:
                        tp += 1
                        right_prediction = True
                    else:
                        fn += 1
            else:
                if len(prediction_value) != n_class_for_predictions:
                    print("len(prediction_value) != n_class_for_predictions")
                    continue

                cell_type_predictions_dict[cell_type_code].append(prediction_value[cell_type_code])

                if len(np.where(prediction_value < 0.5)[0]) == len(prediction_value):
                    # means all values are under the 0.5 threshold, and the cell doesn't belong to any cell type
                    none_of_cell_types += 1
                if len(np.where(prediction_value >= 0.5)[0]) >= 2:
                    # means at least 2 values are over the 0.5 threshold,
                    # and the cell belong to more than one cell type
                    more_than_one_cell_type += 1

                # we want to use sklearn metrics for multiclass metrics
                y_true.append(cell_type_from_code_dict[cell_type_code])
                predicted_cell_type_code = np.argmax(prediction_value)
                y_pred.append(cell_type_from_code_dict[predicted_cell_type_code])
                if cell_type_code == predicted_cell_type_code:
                    right_prediction = True
                # if np.argmax(prediction_value) == cell_type_code:
                #     tp += 1
                # else:
                #     fp += 1
            print(f"Cell {segment[0]} [{cinac_recording.cell_type.strip().lower()}] from {identifier}, predictions: "
                  f"{prediction_value} {right_prediction}")
    if n_class_for_predictions == 1:
        # metrics:
        if (tp + fn) > 0:
            sensitivity = tp / (tp + fn)
        else:
            sensitivity = 1

        if (tn + fp) > 0:
            specificity = tn / (tn + fp)
        else:
            specificity = 1

        if (tp + tn + fp + fn) > 0:
            accuracy = (tp + tn) / (tp + tn + fp + fn)
        else:
            accuracy = 1

        if (tp + fp) > 0:
            ppv = tp / (tp + fp)
        else:
            ppv = 1

        if (tn + fn) > 0:
            npv = tn / (tn + fn)
        else:
            npv = 1

    print("")
    print("-" * 100)
    metrics_title_str = "METRICS for "
    for cell_type_code, n_cells_by_type in n_cells_by_type_dict.items():
        cell_type_str = cell_type_from_code_dict[cell_type_code]
        metrics_title_str = metrics_title_str + f"{n_cells_by_type} {cell_type_str}, "
    print(f"{metrics_title_str[:-2]}")
    # print(f"METRICS for {n_pyr_cell} pyramidal cells and {n_in_cell} interneurons")
    if n_class_for_predictions == 1:
        print("-" * 100)
        print(f"Accuracy {accuracy}")
        print(f"Sensitivity {sensitivity}")
        print(f"Specificity {specificity}")
        print(f"PPV {ppv}")
        print(f"NPV {npv}")
        print("-" * 100)
    else:
        print(f"N cell classified has none of the cell types: {none_of_cell_types}")
        print(f"N cell classified has more than one cell type: {more_than_one_cell_type}")
        # Print the confusion matrix
        print("--- confusion matrix ---")
        print(sklearn.metrics.confusion_matrix(y_true, y_pred))

        # Print the precision and recall, among other metrics
        print(sklearn.metrics.classification_report(y_true, y_pred, digits=3))

    data_hist_dict = dict()
    for cell_type_code, cell_type_predictions in cell_type_predictions_dict.items():
        cell_type_str = cell_type_from_code_dict[cell_type_code]
        data_hist_dict[cell_type_str] = cell_type_predictions
        print(f"For {cell_type_str}, mean {np.round(np.mean(cell_type_predictions), 2)}, "
              f"std {np.round(np.std(cell_type_predictions), 2)}, "
              f"min {np.round(np.min(cell_type_predictions), 2)}, "
              f"max {np.round(np.max(cell_type_predictions), 2)}")
        if save_cell_type_distribution:
            plot_hist_distribution(distribution_data=[p * 100 for p in cell_type_predictions],
                                   description=f"hist_prediction_distribution_{cell_type_str}",
                                   path_results=results_path,
                                   tight_x_range=False,
                                   twice_more_bins=True,
                                   xlabel=f"{cell_type_str}",
                                   save_formats="png")
    print("-" * 100)
